import torch.nn as nn
from torch import log
import numpy as np

from .utils_cage import probability


def log_likelihood_loss_supervised(theta, pi, y, m, s, k, n_classes, continuous_mask, qc, device):
	'''
		Joint Learning utils: Negative log likelihood loss, used in loss 4 in :cite:p:`DBLP:journals/corr/abs-2008-09887`

	Args:
		theta: [n_classes, n_lfs], the parameters
		pi: [n_classes, n_lfs], the parameters
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		s: [n_instances, n_lfs], s[i][j] is the continuous score of ith instance given by jth continuous LF
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		continuous_mask: [n_lfs], continuous_mask[i] is 1 if ith LF has continuous counter part, else it is 0
		qc: a float value OR [n_lfs], qc[i] quality index for ith LF
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a real value, summation over (the log of probability for an instance)
	'''
	prob = probability(theta, pi, m, s, k, n_classes, continuous_mask, qc, device)
	prob = (prob.t() / prob.sum(1)).t()
	return nn.NLLLoss()(log(prob), y)

def entropy(probabilities):
	'''
		Joint Learning utils: Entropy, Used in loss 2 in :cite:p:`DBLP:journals/corr/abs-2008-09887`

	Args:
		probabilities: [num_unsup_instances, num_classes], probabilities[i][j] is probability of ith instance being jth class
	
	Return:
		a real value, the entropy value of given probability
	'''
	entropy = - (probabilities * log(probabilities)).sum(1)
	return entropy.sum() / probabilities.shape[0]

def kl_divergence(probs_p, probs_q):
	'''
		Joint Learning utils: KL divergence of two probabilities, used in loss 6 in :cite:p:`DBLP:journals/corr/abs-2008-09887`
		
	Args:
		probs_p: [num_instances, num_classes]
		probs_q: [num_instances, num_classes]
	
	Return:
		a real value, the KL divergence of given probabilities
	'''
	return (probs_p * log(probs_p / probs_q)).sum() / probs_p.shape[0]


def find_indices(data, data_sub):
	'''
		A helper function for subset selection

	Args:
		data: the complete data, torch tensor of shape [num_instances, num_classes]
		data_sub: the subset of 'data' whose indices are to be found. Should be of same shape as 'data'
	
	Return:
		list of indices, to be found from the result of apricot library
	'''
	indices = []
	for element in data_sub:
		x = np.where((data.cpu().numpy() == element.cpu().numpy()).all(axis=1))[0]
		indices.append(x[0])
	return indices


def get_similarity_kernel(preds):
	'''
		A helper function for subset selection

	Args:
		preds: numpy.ndarray of shape (num_samples,)
	
	Return:
		numpy.ndarray of shape (num_sample, num_samples)

	'''
	num_samples = len(preds)
	kernel_matrix = np.zeros((num_samples, num_samples))
	for pred in np.unique(preds):
		x = np.where(preds == pred)[0]
		prod = np.transpose([np.tile(x, len(x)), np.repeat(x, len(x))])
		kernel_matrix[prod] = 1
	return kernel_matrix



import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np

def exp_term_for_constraints(rule_classes, num_classes, C):
	'''
	Func Desc:
	Compute the exponential term for the constraints

	Input:
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C

	Output:
	the required exponential term
	'''
	rule_classes_tensor = tf.to_float(tf.convert_to_tensor(rule_classes))
	#rule_classes_tensor = tf.reshape(rule_classes_tensor,[1,rule_classes])
	rule_classes_tensor = tf.expand_dims(rule_classes_tensor,0)
	class_types_tensor = tf.to_float(tf.convert_to_tensor(np.arange(num_classes).reshape(num_classes,1)))	
	#[num_classes,num_rules]
	class_rule_constraint = tf.to_float(tf.equal(class_types_tensor,rule_classes_tensor)) - 1.0
	class_rule_constraint = tf.exp(C*class_rule_constraint)
	return class_rule_constraint


def pr_product_term(weights, rule_classes, num_classes, C):
	'''
	Func Desc:
	Compute the probability product term for the constraints

	Input:
	weights ([batch_size, num_rules]) - the w_network weights
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C

	Output:
	the required product term
	'''
	# weights: [batch_size, num_rules]
	class_rule_constraint = exp_term_for_constraints(rule_classes, num_classes, C)
	#class_rule_constraint = tf.Print(class_rule_constraint,[tf.shape(class_rule_constraint)],message="class_rule_constraint")
	#[num_classes,1,num_rules]
	class_rule_constraint = tf.expand_dims(class_rule_constraint,axis=1)
	#[1,batch_size,num_rules]
	weights = tf.expand_dims(weights,axis=0) 	 
	# [num_classes,batch_size,num_rules]
	t1 = class_rule_constraint * weights
	# [1, batch_size, num_rules]
	t2 = 1-weights
	#[num_classes,batch_size,num_rules]
	t = t1+t2
	#t = tf.Print(t, [t,tf.shape(t)],message="t and shape of t")
	product_term = tf.reduce_prod(t,axis=-1)
	#[batch_size, num_classes]
	product_term = tf.transpose(product_term)
	return product_term


def get_q_y_from_p(f_probs, weights, rule_classes, num_classes, C):
	'''
	Func Desc:
	Compute the q_y term from the p (f_network) distribution 

	Input:
	f_probs ([batch_size, num_classes])
	weights ([batch_size, num_rules]) - the w_network weights
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C

	Output:
	the required q_y term
	'''
	# f_probs: [batch_size, num_classes]
	# weights: [batch_size, num_rules]
	psi = 1e-20
	product_term = pr_product_term(weights, rule_classes, num_classes, C)
	result = f_probs * product_term
	normalizer = tf.reduce_sum(result,axis=-1,keepdims=True)
	result = result/(normalizer + psi)
	return result

def get_q_r_from_p(f_probs, weights, rule_classes, num_classes, C):
	'''
	Func Desc:
	Compute the q_r term from the p (f_network) distribution 

	Input:
	f_probs ([batch_size, num_classes])
	weights ([batch_size, num_rules]) - the w_network weights
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C

	Output:
	the required q_r term
	'''
	# f_probs: [batch_size, num_classes]
	# weights: [batch_size, num_rules]
	psi = 1e-20
	#[batch_size, num_classes]
	pr_product_t = pr_product_term(weights, rule_classes, num_classes, C)
	#[batch_size, 1, num_classes] 
	product_term = tf.expand_dims(pr_product_t,axis=1)
	#[num_rules, num_classes]
	class_rule_constraint = tf.transpose(exp_term_for_constraints(rule_classes, num_classes, C))
	#[1, num_rules, num_classes]
	class_rule_constraint = tf.expand_dims(class_rule_constraint, axis=0)
	#[batch_size, num_rules, 1]
	w = tf.expand_dims(weights,2)
	#[batch_size, num_rules, num_classes]
	divisior = w*class_rule_constraint + (1-w)
	product_term = product_term / (divisior + psi)
	#[batch_size, 1, num_classes]
	f_probs = tf.expand_dims(f_probs,axis=1)
	#[batch_size, num_rules, num_classes]
	product_term = product_term * f_probs * class_rule_constraint
	#[batch_size, num_rules]
	sum_over_y_term = tf.reduce_sum(product_term,axis=-1)
	prob_q_r_eq_1 = weights * sum_over_y_term


	prob_q_r_eq_0 = tf.reduce_sum(f_probs * product_term, axis=-1)
	prob_q_r_eq_0 = (1 - weights) * prob_q_r_eq_0

	prob_q_r_eq_1 = prob_q_r_eq_1 / (prob_q_r_eq_0 + prob_q_r_eq_1)

	return prob_q_r_eq_1


def theta_term_in_pr_loss(f_logits, f_probs, weights, rule_classes, num_classes, C, d):
	'''
	Func Desc:
	Compute the theta term in the pr loss 

	Input:
	f_logits ([batch_size, num_classes])
	f_probs ([batch_size, num_classes])
	weights ([batch_size, num_rules]) - the w_network weights
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C
	d ([batch_size,1])

	Output:
	the required theta term (third term in equation 14) -  used to supervise f (classification) network from instances in U
	'''
	#[batch_size, num_classes]
	q_y = get_q_y_from_p(f_probs, weights, rule_classes, num_classes, C)
	cross_ent_q_p = tf.nn.softmax_cross_entropy_with_logits(labels=q_y,logits=f_logits)
	cross_ent_q_p = (1-d) * cross_ent_q_p #defined only for instances in U, so mask by (1-d)
	result = tf.reduce_mean(cross_ent_q_p)
	return result

def phi_term_in_pr_loss(m, w_logits, f_probs, weights, rule_classes, num_classes, C, d):
	'''
	Func Desc:
	Compute the phi term in the pr loss

	Input:
	m ([batch_size, num_rules]) - mij = 1 if ith example is associated with jth rule 
	w_logits ([batch_size, num_rules])
	f_probs ([batch_size, num_classes])
	weights ([batch_size, num_rules]) - the w_network weights
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C
	d ([batch_size,1])

	Output:
	the required phi term (fourth term in equation 14) - used to superwise w (rule) network from instances in U
	'''
	#w_logits: [batch_size, num_rules]
	#m: [batch_size, num_rules]
	psi = 1e-20
	q_r_1 = get_q_r_from_p(f_probs, weights, rule_classes, num_classes, C)
	#[batch_size, num_rules]
	cross_ent_q_w = tf.nn.sigmoid_cross_entropy_with_logits(labels=q_r_1, logits=w_logits)
	cross_ent_q_w = tf.reduce_sum(cross_ent_q_w*m,axis=-1)
	#normalizer_cross_ent_q_w = tf.reduce_sum(m,axis=-1)
	#cross_ent_q_w = cross_ent_q_w / (normalizer_cross_ent_q_w + psi)
	cross_ent_q_w = cross_ent_q_w * (1-d)
	cross_ent_q_w = tf.reduce_mean(cross_ent_q_w)
	return cross_ent_q_w

def pr_loss(m, f_logits, w_logits, f_probs, weights, rule_classes, num_classes, C, d):
	'''
	Func Desc:
	Compute the  pr loss 

	Input:
	m ([batch_size, num_rules]) - mij = 1 if ith example is associated with jth rule 
	f_logits 
	w_logits ([batch_size, num_rules]) - logit before sigmoid activation in w network
	f_probs ([batch_size, num_classes]) - output of f network
	weights ([batch_size, num_rules]) - the sigmoid output from w network
	rule_classes ([num_rules,1]) - a list of classes associated with the rules
	num_classes (int)
	C - lamda in equation 10 (hyperparameter)
	d ([batch_size,1]) - if ith instance is from "d" set (labelled data) d[i] = 1, else if ith instance is from "U" set, d[i] = 0

	Output:
	the required phi term
	'''

	#theta_term : (third term in equation 14) (used to supervise f (classification) network from instances in U )
	#phi term : (fourth term in equation 14)  (used to superwise w (rule) network from instances in U )

	# m : rule_firing matrix: [batch_size, num_rules]
	# w_logits: logit before sigmoid activation in w network: [batch_size, num_rules]
	# weights: sigmoid output from w network: [batch_size, num_rules]
	# f_probs: output of f network: [batch_size, num_classes]
	# C: \lamda in equation 10 (hyperparameter)
	# d : [batch_size], d[i] = 0 if ith instance is from "U" set, 1 if ith instance is from "d" set (labeled data)

	theta_term = theta_term_in_pr_loss(f_logits, f_probs, weights, rule_classes, num_classes, C, d)
	cross_ent_q_w = phi_term_in_pr_loss(m, w_logits, f_probs, weights, rule_classes, num_classes, C, d)
	return theta_term + cross_ent_q_w







# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import functools
import math
import numpy as np
import sys,os

# from .my_checkpoints import MRUCheckpoint, CheckpointsFactory
# from .my_data_types import *
# # import my_gen_cross_entropy_utils as gcross_utils
# from .my_gen_cross_entropy_utils import *
# from .my_gen_cross_entropy_utils import *
# # import my_pr_utils
# from .my_pr_utils import *
# from .my_test import HLSTest
# from .my_train import HLSTrain
# from .my_utils import print_tf_global_variables, updated_theta_copy

from .checkpoints import MRUCheckpoint, CheckpointsFactory
from .data_types import *
# import gen_cross_entropy_utils as gcross_utils
from .gen_cross_entropy_utils import *
from .gen_cross_entropy_utils import *
# import pr_utils
from .pr_utils import *
from .test import HLSTest
from .train import HLSTrain
from .utils import print_tf_global_variables, updated_theta_copy



class HighLevelSupervisionNetwork:
	# Parameters
	display_step = 1
	'''
	Class Desc:
	Initialize HLS with number of input features, number of classes, number of rules and the f and the w network.
	f network is the classification network (P_{theta})
	w network is the rule network (P_{j,phi})
	'''
	def __init__(self, num_features, num_classes, num_rules,
			num_rules_to_train, rule_classes,
			w_network, f_network, 
			f_d_epochs, f_d_U_epochs, f_d_adam_lr, f_d_U_adam_lr, dropout_keep_prob, 
			f_d_metrics_pickle, f_d_U_metrics_pickle, early_stopping_p, f_d_primary_metric, mode, data_dir, 
			tensorboard_dir, checkpoint_dir, checkpoint_load_mode, gamma, lamda, raw_d_x=None, raw_d_L=None):
		'''
		Func Desc:
		initializes the class member variables with the provided arguments

		Input:
		self
		num_features
		num_classes
		num_rules
		num_rules_to_train
		rule_classes
		w_network
		f_network
		raw_d_x (default = None)
		raw_d_L (default = None)

		Output:

		'''
		
		# Modules for testing/training

		self.mode = mode
		self.gamma = gamma
		self.lamda = lamda
		self.raw_d_x = raw_d_x
		self.raw_d_L = raw_d_L
		self.rule_classes_list = rule_classes
		self.rule_classes = tf.convert_to_tensor(rule_classes)
		self.num_features = num_features
		self.num_classes = num_classes
		self.num_rules = num_rules
		self.num_rules_to_train = num_rules_to_train
		self.w_var_scope = 'w_network'
		self.f_var_scope = 'f_network'
		self.w_network = functools.partial(w_network, self.w_var_scope,
				self.num_rules)
		self.f_network = functools.partial(f_network, self.f_var_scope)
		self.parse_params(f_d_epochs, f_d_U_epochs, f_d_adam_lr, f_d_U_adam_lr)

		# Save global step for each different kind of run
		self.global_steps = {}
		# Global global step 
		self.global_step = tf.train.get_or_create_global_step()

		# Create the compute graphs
		# dropout rate used in f and w networks
		self.dropout_keep_prob = tf.placeholder(tf.float32,name="keep_prob")
		self.dropout_train_dict = {self.dropout_keep_prob: dropout_keep_prob}
		self.dropout_test_dict = {self.dropout_keep_prob: 1.0}
		
		self.train = HLSTrain(self, f_d_metrics_pickle, 
									f_d_U_metrics_pickle, 
									f_d_adam_lr, 
									f_d_U_adam_lr, 
									early_stopping_p, 
									f_d_primary_metric, 
									mode, 
									data_dir)
		self.test = HLSTest(self)
		
		self.make_f_d_train_ops()
		self.make_f_d_U_train_ops()

		# Print all global variables
		print_tf_global_variables()

		# Initialize all variables
		self.init = tf.global_variables_initializer()
		sess_config = tf.ConfigProto()
		sess_config.gpu_options.allow_growth = True
		self.sess = tf.Session(config=sess_config)

		self.writer = tf.summary.FileWriter(tensorboard_dir, self.sess.graph)
		# Now that all variables and the session is created, create a
		# checkpoint saver. We use a single saver for all variables
		self.mru_saver = MRUCheckpoint(checkpoint_dir, self.sess, tf.global_variables())
		self.best_savers = CheckpointsFactory(self.sess, self.global_steps)
		feed_dict = {}
		self.sess.run(self.init, feed_dict=feed_dict)
		if checkpoint_load_mode == 'mru':
			self.restored = self.mru_saver.restore_if_checkpoint_exists()
		else:
			saver = self.best_savers.get_best_saver(checkpoint_load_mode)
			self.restored = saver.restore_best_checkpoint_if_exists()
			if not self.restored:
				raise ValueError('Asked to restore best checkpoint of %s but not previously checkpointed' %
						checkpoint_load_mode)

	def parse_params(self, f_d_epochs, f_d_U_epochs, f_d_adam_lr, f_d_U_adam_lr):
		'''
		Func Desc:
		Parses the given parameters 

		Input:
		self 

		Sets:
		f_d_epochs
		f_d_U_epochs
		initial_f_d_adam_lr
		initial_f_d_U_adam_lr
		'''
		self.f_d_epochs     =   f_d_epochs
		self.f_d_U_epochs   = f_d_U_epochs
		self.initial_f_d_adam_lr = f_d_adam_lr
		self.initial_f_d_U_adam_lr = f_d_U_adam_lr

	# Create the train op for training with d only
	def make_f_d_train_ops(self):
		'''
		Func Desc:
		create the train_ops based on labelled data only

		Input:
		self

		Output:

		'''
		self.f_d_global_step = tf.Variable(0, trainable=False, name='f_d_global_step')
		inc_f_d_global_step = tf.assign_add(self.f_d_global_step, 1)
		self.global_steps[f_d] = self.f_d_global_step
		self.f_d_adam_lr = tf.placeholder(tf.float32,name='f_d_U_adam_lr')

		# [batch size, features]
		self.f_x = tf.placeholder(tf.float32, shape=[None, self.num_features],
				name='f_d_x')
		self.f_d_labels = tf.placeholder(tf.float32, shape=[None,
			self.num_classes], name='f_d_labels')

		f_dict = {'x': self.f_x, 'labels': self.f_d_labels}

		self.f_d_logits = self.f_network(f_dict, self.num_classes,
				dropout_keep_prob=self.dropout_keep_prob)

		self.f_d_probs = tf.math.softmax(self.f_d_logits, axis=-1)
		self.f_d_preds = tf.argmax(self.f_d_probs, axis=-1)        

		
		model_loss = tf.nn.softmax_cross_entropy_with_logits(
			labels=self.f_d_labels, logits=self.f_d_logits)

		self.f_d_loss = tf.reduce_mean(model_loss)

		self.f_d_optimizer = tf.train.AdamOptimizer(
				learning_rate=self.f_d_adam_lr,
				name='f_d_Adam')

		with tf.control_dependencies([inc_f_d_global_step]):
			self.f_d_train_op = self.f_d_optimizer.minimize(self.f_d_loss, global_step=self.global_step)

	def make_f_d_U_train_ops(self):
		self.f_d_U_global_step = tf.Variable(0, trainable=False, name='f_d_U_global_step')
		inc_f_d_U_global_step = tf.assign_add(self.f_d_U_global_step, 1)
		self.global_steps[f_d_U] = self.f_d_U_global_step
		'''
		Func desc:
		make_f_d_U_train_ops i.e. training ops by combining labelled and unlabelled data, compute the training objective and aim to minimize the loss function using the adam optimizer

		Input:
		self object

		Sets:
		* x : feature representation of instance
			- shape : [batch_size, num_features]

		* l : Labels assigned by rules
			- shape [batch_size, num_rules]
			- l[i][j] provides the class label provided by jth rule on ith instance
			- if jth rule does not fire on ith instance, then l[i][j] = num_classes (convention)
			- in snorkel, convention is to keep l[i][j] = -1, if jth rule doesn't cover ith instance
			- class labels belong to {0, 1, 2, .. num_classes-1}

		* m : Rule coverage mask
			- A binary matrix of shape [batch_size, num_rules]
			- m[i][j] = 1 if jth rule cover ith instance
			- m[i][j] = 0 otherwise

		* L : Instance labels
			- shape : [batch_size, 1]
			- L[i] = label of ith instance, if label is available i.e. if instance is from labeled set d
			- Else, L[i] = num_clases if instances comes from the unlabeled set U
			- class labels belong to {0, 1, 2, .. num_classes-1}

		* d : binary matrix of shape [batch_size, 1]
			- d[i] = 1 if instance belongs to labeled data (d), d[i]=0 otherwise
			- d[i]=1 for all instances is from d_processed.p
			- d[i]=0 for all instances in other 3 pickles {U,validation,test}_processed.p
		
		* r : A binary matrix of shape [batch_size, num_rules]
			- r[i][j]=1 if jth rule was associated with ith instance
			- Highly sparse matrix
			- r is a 0 matrix for all the pickles except d_processed.p
			- Note that this is different from rule coverage matrix "m"
			- This matrix defines the rule,example pairs provided as supervision 

		* s : A similarity measure matrix shape [batch_size, num_rules]
			- s[i][j] is in [0,1]

		* n : A vector of size [num_rules,]
			- Mask for s (denotes whether particular rule is continuous or discrete)

		* k : a vector of size [num_rules,]
			- #LF classes ie., what class each LF correspond to, range: 0 to num_classes-1

		Computes:
		weights, w_logits of rule network ([batch_size, num_rules]) - Used to train P_j_phi(r_j/x_i) i.e. whether rij = 1 for the ith instance and jth rule 
		f_logits of the classification network - Used to train P_j_theta(l_j/x_i) i.e. the probability of ith instance belonging to jth class
		LL_phi term
		LL_theta term
		Training objective term
		Minimum loss using adam optimizer
		'''

		self.f_d_U_adam_lr = tf.placeholder(tf.float32,name='f_d_U_adam_lr')
		self.f_d_U_x = tf.placeholder(
				tf.float32,
				shape=[None, self.num_features],
				name='f_d_U_x')
		# l
		self.f_d_U_l = tf.placeholder(
				tf.int32,
				shape=[None, self.num_rules],
				name='f_d_U_l')
		# m 
		self.f_d_U_m = tf.placeholder(tf.float32, shape=[None,
			self.num_rules], name='f_d_U_m')
		# L
		L = self.f_d_U_L = tf.placeholder(tf.int32, shape=[None, 1], name='f_d_U_L')
		# d
		d = self.f_d_U_d = tf.placeholder(tf.float32, shape=[None, 1], name='f_d_U_d')
		L = tf.squeeze(L)
		d = tf.squeeze(d)

		r = self.f_d_U_r = tf.placeholder(tf.float32, shape=[None,self.num_rules], name='f_d_U_r')

		#weights: [batch_size, num_rules]
		#w_logits: [batch_size, num_rules]
		weights, w_logits = self.get_weights_and_logits_f_d_U(self.f_d_U_x)
		self.f_d_U_weights = weights
		self.f_d_U_num_d = tf.reduce_sum(d) #number of labeled instances in a batch

		# w_network computation is done. Now run f_network to get logits for
		# this batch
		f_dict = {'x': self.f_d_U_x}
		f_logits = self.f_network(f_dict, self.num_classes, reuse=True, 
								dropout_keep_prob=self.dropout_keep_prob)
		self.f_d_U_probs = tf.math.softmax(f_logits, axis=-1)
		self.f_d_U_preds = tf.argmax(self.f_d_U_probs, axis=-1)
		self.joint_f_w_score = self.joint_scores_from_f_and_w(self.f_d_U_weights,self.f_d_U_m,self.f_d_U_probs)

		# Do this so that the cross-entropy does not blow for data from U
		# The actual value of cross-entropy for U does not matter since it
		# will be multiplied by 0 anyway.        
		L = L % self.num_classes

		# Ok now compute the loss LL_theta which is on d data
		L_one_hot = tf.one_hot(L, self.num_classes)
		LL_theta = tf.nn.softmax_cross_entropy_with_logits(logits=f_logits,
				labels=L_one_hot)
		LL_theta = d * LL_theta
		LL_theta = tf.reduce_mean(LL_theta) # loss of f network on labeled data d

		LL_theta = LL_theta # loss of f network on labeled data d
								 # first term in eqn 5 (LL(\theta))


		# LL(\phi) term
		LL_phi = self.compute_LL_phi(w_logits=w_logits, 
														  weights=self.f_d_U_weights,
														  l=self.f_d_U_l, 
														  m=self.f_d_U_m,
														  L=L,
														  d=d,
														  r=self.f_d_U_r)
		

		f_cross_training_optimizer = tf.train.AdamOptimizer(
				learning_rate=self.f_d_U_adam_lr,
				name='f_d_U_Adam')

		training_var_list = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES)


		if 'implication' == self.mode:
			implication_loss = self.implication_loss(weights=self.f_d_U_weights,
													f_probs=self.f_d_U_probs,
													m=self.f_d_U_m,
													rule_classes=self.rule_classes,
													num_classes=self.num_classes,
													d=d)
			
			self.f_d_U_implication_loss = LL_phi \
										+ LL_theta \
										+ self.gamma*implication_loss

			with tf.control_dependencies([inc_f_d_U_global_step,  ]):
				self.f_d_U_implication_op = f_cross_training_optimizer.minimize(
						self.f_d_U_implication_loss,
						var_list=training_var_list)

		if 'pr_loss' == self.mode:
# 			pr_loss = my_pr_utils.pr_loss(m=self.f_d_U_m,
			pr_loss_val = pr_loss(m=self.f_d_U_m,
									f_logits=f_logits, 
									w_logits=w_logits, 
									f_probs=self.f_d_U_probs,
									weights=self.f_d_U_weights,
									rule_classes=self.rule_classes,
									num_classes=self.num_classes, 
									C=0.1,
									d=d)
			self.pr_loss = LL_theta + LL_phi + self.gamma*pr_loss_val 
			with tf.control_dependencies([inc_f_d_U_global_step,  ]):
				self.pr_train_op = f_cross_training_optimizer.minimize(
									self.pr_loss,
									var_list=training_var_list)

		if 'gcross' == self.mode:
			self.f_d_U_snork_L = tf.placeholder(
					tf.float32,
					shape=[None, self.num_classes],
					name='f_d_U_snork_L')

			loss_on_d = LL_theta
# 			loss_on_U = gcross_utils.generalized_cross_entropy(f_logits,self.f_d_U_snork_L,
# 													self.lamda)
			loss_on_U = generalized_cross_entropy(f_logits,self.f_d_U_snork_L,
													self.lamda)
			self.gcross_loss = loss_on_d + self.gamma*loss_on_U
			with tf.control_dependencies([inc_f_d_U_global_step,  ]):
				self.gcross_train_op = f_cross_training_optimizer.minimize(
									self.gcross_loss,
									var_list=training_var_list)

		if 'gcross_snorkel' == self.mode:
			self.f_d_U_snork_L = tf.placeholder(
					tf.float32,
					shape=[None, self.num_classes],
					name='f_d_U_snork_L')

			loss_on_d = LL_theta
# 			loss_on_U = gcross_utils.generalized_cross_entropy(f_logits,self.f_d_U_snork_L,
# 													self.lamda)
			loss_on_U = generalized_cross_entropy(f_logits,self.f_d_U_snork_L,
													self.lamda)
			self.snork_gcross_loss = loss_on_d + self.gamma*loss_on_U
			#self.snork_gcross_loss = loss_on_d + loss_on_U
			with tf.control_dependencies([inc_f_d_U_global_step,  ]):
				self.snork_gcross_train_op = f_cross_training_optimizer.minimize(
									self.snork_gcross_loss,
									var_list=training_var_list)

		if 'label_snorkel' == self.mode or 'pure_snorkel' == self.mode:
			self.f_d_U_snork_L = tf.placeholder(
					tf.float32,
					shape=[None, self.num_classes],
					name='f_d_U_snork_L')
			loss_on_d = LL_theta
			self.pure_snorkel_loss = tf.nn.softmax_cross_entropy_with_logits(
									 labels=self.f_d_U_snork_L,logits=f_logits)
			self.pure_snorkel_loss = tf.reduce_mean(self.pure_snorkel_loss)
			self.label_snorkel_loss = loss_on_d + self.gamma*self.pure_snorkel_loss

			if 'label_snorkel' == self.mode:
				with tf.control_dependencies([inc_f_d_U_global_step,  ]):
					self.label_snorkel_train_op = f_cross_training_optimizer.minimize(
										self.label_snorkel_loss,
										var_list=training_var_list)

			if 'pure_snorkel' == self.mode:
				with tf.control_dependencies([inc_f_d_U_global_step,  ]):
					self.pure_snorkel_train_op = f_cross_training_optimizer.minimize(
										self.pure_snorkel_loss,
										var_list=training_var_list)

		if 'learn2reweight' == self.mode:
			len_raw_d_x = len(self.raw_d_x)
			raw_d_bs = min(len_raw_d_x,32)
			raw_d_x = tf.get_variable(name="raw_d_x", initializer=self.raw_d_x, trainable=False)
			raw_d_x = tf.to_float(raw_d_x)
			raw_d_L = tf.get_variable(name="raw_d_L", initializer=self.raw_d_L, trainable=False)
			raw_d_L = tf.to_int32(raw_d_L)
			#raw_d_L = tf.expand_dims(raw_d_L,1)
			batch_points = tf.random.uniform([raw_d_bs],minval=0,maxval=len_raw_d_x, dtype=tf.int32)
			one_hot_batch_points_float = tf.one_hot(batch_points,len_raw_d_x,dtype=tf.float32)
			batch_raw_d_x = tf.matmul(one_hot_batch_points_float,raw_d_x)
			one_hot_batch_points_int = tf.one_hot(batch_points,len_raw_d_x,dtype=tf.int32)
			batch_raw_d_L = tf.matmul(one_hot_batch_points_int,raw_d_L)
			batch_raw_d_L = tf.squeeze(batch_raw_d_L) 

			self.f_d_U_snork_L = tf.placeholder(
					tf.float32,
					shape=[None, self.num_classes],
					name='f_d_U_snork_L')

			# 1. initialize epsilon
			# [batch_size]
			epsilon = tf.zeros(tf.shape(self.f_d_U_x)[0])
			
			# 2. compute epsilon weighted loss (ewl) for batch
			#[batch_size, num_classes]
			f_logits = self.f_network(f_dict, self.num_classes, reuse=True,
				dropout_keep_prob=self.dropout_keep_prob)
			#[batch_size]
			unweighted_loss = tf.nn.softmax_cross_entropy_with_logits(labels=self.f_d_U_snork_L,logits=f_logits)
			weighted_loss = epsilon * unweighted_loss 
			weighted_loss = tf.reduce_sum(weighted_loss)
			
			# 3. compute grads of ewl wrt thetas
			thetas = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.f_var_scope)
			grads_thetas = tf.gradients(ys=weighted_loss,xs=thetas,stop_gradients=epsilon)

			
			# 4. update theta
			theta_hat = updated_theta_copy(
													grads=grads_thetas,
													variables=thetas,
													lr=self.lamda,
													mode=-1)

			# 5. compute unweighted loss on raw_d with updated theta (theta_hat)            
			f_dict_on_d = {'x': batch_raw_d_x}
			f_logits_on_d = self.f_network(f_dict_on_d, self.num_classes, 
											   reuse=False, ph_vars=theta_hat,
											   dropout_keep_prob=self.dropout_keep_prob)
			raw_d_L_one_hot = tf.one_hot(batch_raw_d_L,self.num_classes,dtype=tf.float32)
			unweighted_loss = tf.nn.softmax_cross_entropy_with_logits(labels=raw_d_L_one_hot,
																		  logits=f_logits_on_d)
			unweighted_loss = tf.reduce_mean(unweighted_loss)

			# 6. compute grads of unweighted loss wrt epsilons
			grad_epsilon = tf.gradients(ys=unweighted_loss,xs=epsilon,stop_gradients=thetas)[0]
			#grad_epsilon = tf.Print(grad_epsilon,[grad_epsilon],message="\n\n\n grad_epsilon \n\n\n")                

			# 7. truncate and normalize grad-epsilons to get w
			w_tilde = tf.nn.relu(-grad_epsilon)
			w_norm = w_tilde/(tf.reduce_sum(w_tilde) + 1e-25)
			#w_norm = tf.Print(w_norm,[w_norm],message="\n\n\n w_norm \n\n\n")

			# 8. Compute ewl for batch with original theta and normalized weights
			f_logits = self.f_network(f_dict,self.num_classes,reuse=True,
									  dropout_keep_prob=self.dropout_keep_prob)
			unweighted_loss = tf.nn.softmax_cross_entropy_with_logits(
									 labels=self.f_d_U_snork_L,logits=f_logits)
			w_norm = tf.stop_gradient(w_norm)
			weighted_loss = w_norm * unweighted_loss
			weighted_loss = tf.reduce_sum(weighted_loss)
			self.l2r_loss = weighted_loss
			with tf.control_dependencies([inc_f_d_U_global_step,  ]):
				# 9. Compute grads of ewl wrt to original theta to obtain Update theta operation
				self.l2r_train_op = f_cross_training_optimizer.minimize(
									self.l2r_loss,
									var_list=thetas)

	def compute_LL_phi(self, w_logits, weights, l, m, L, d, r):
		'''
		Func desc: 
		Computes the LL_phi term coming in the training objective

		Input:
		self object
		w_logits([batch_size, num_rules]) - 
		weights([batch_size, num_rules]) - the weights matrix corresponding to rule network(w_network) in the algorithm
		l([batch_size, num_rules]) - labels assigned by the rules
		m([batch_size, num_rules]) - the rule coverage matrix where m_ij = 1 if jth rule covers ith instance 
		L([batch_size, 1]) - L_i = 1 if the ith instance has already a label assigned to it in the dataset
		d([batch_size, 1]) - d_i = 1 if the ith instance is from labelled dataset
		r([batch_size, num_rules]) - the rule association matrix where r_ij = 1 if jth rule is associated with ith instance (r_ij = 1 => m_ij = 1)
		
		Output:
		loss(real number > 0) - the value of the LL_phi term
		'''
		psi = 1e-25
		L = tf.expand_dims(L,1)
		# [batch_size, num_rules]
		L = tf.tile(L,[1,self.num_rules])
		loss = tf.nn.sigmoid_cross_entropy_with_logits(labels=tf.to_float(tf.equal(l,L)),
													   logits=w_logits)
		loss = m*loss
		loss = (tf.to_float(tf.not_equal(l,L)) * loss) + (tf.to_float(tf.equal(l,L)) * r * loss)
# 		gcross_loss = gcross_utils.generalized_cross_entropy_bernoulli(weights,0.2)
		gcross_loss = generalized_cross_entropy_bernoulli(weights,0.2)
		gcross_loss = gcross_loss * m * tf.to_float(tf.equal(l,L)) * (1-r)
		loss = loss + gcross_loss
		loss = tf.reduce_sum(loss,axis=-1)
		loss = loss * d
		loss = tf.reduce_mean(loss)
		return loss


	def implication_loss(self, weights, f_probs, m, rule_classes, num_classes, d):
		''' 
		Func desc:
		Computes the implication loss value

		input:
		self object
		weights([batch_size, num_rules]) - the weights matrix corresponding to rule network(w_network) in the algorithm
		f_probs([batch_size, 1]) - the prob values from classification network (f_network)
		m([batch_size, num_rules]) - the rule coverage matrix where m_ij = 1 if jth rule covers ith instance
		rule_classes
		num_classes(non_negative integer) - number of available classes
		d([batch_size, 1]) - d_i = 1 if the ith instance is from labelled dataset

		output:
		-obj (real number) - the implication loss value
		'''
		# computes implication loss (Equation 4 in the paper)
		# weights are P_{j,\phi} values from the w network (rule network)
		# weights: [batch_size, num_rules] 
		# f_probs are probabilities from the f network (classification network)
		# f_probs: [batch_size, num_classes]
		psi = 1e-25 # a small value to avoid nans

		#[num_rules, num_classes]
		one_hot_mask = tf.one_hot(rule_classes,num_classes,dtype=tf.float32)
		#[batch_size, num_rules]
		f_probs = tf.matmul(f_probs, one_hot_mask, transpose_b=True)
		obj = 1 - (weights * (1 - f_probs)) #(Argument of log in equation 4)

		# computing last term of equation 5, will multiply with gamma outside this function
		obj = m*tf.log(obj + psi)
		obj = tf.reduce_sum(obj, axis=-1)
		obj = obj * (1-d) #defined only for instances in U, so mask by (1-d)
		obj = tf.reduce_mean(obj)
		return -obj

	def get_weights_and_logits_f_d_U(self, x):
		'''
		Func desc:
		compute and get the weights and logits for the rule network (w_network) 

		Input: 
		self object
		x([batch_size, num_features]) - instance matrix

		Output:
		weights([batch_size, num_rules]) - the r_ij values i.e. the possibility of a rule overfitting on an instance (r_ij = 0 for ith instance and jth rule)
		w_logits([batch_size, num_rules]) - 
		'''
		# Need to run the w network for each rule for the same x
		#
		# [batch_size, num_rules, num_features]
		x_shape = tf.shape(x)
		mul = tf.convert_to_tensor([1, self.num_rules_to_train])
		expanded_x = tf.tile(x, mul)
		# Need a python integer as the last dimension so that defining neural
		# networks work later. Hence use num_features instead of x_shape[1]
		x = tf.reshape(expanded_x , [x_shape[0], self.num_rules_to_train,
			self.num_features])

		batch_size = x_shape[0]
		rules_int = tf.convert_to_tensor([list(range(0,
			self.num_rules_to_train))])
		# Need to tile rules_int batch_size times
		#
		# tilevar should be a 1-D tensor with number of values equal to number
		# of columns in rules_int. Each column specifies the number of times
		# that axis in rules_int will be replicated.
		#
		# Following will replicate the rows of rules_int batch_size times and
		# leave the columns unchanged
		tilevar = tf.convert_to_tensor([batch_size, 1])
		rules_int = tf.tile(rules_int, tilevar) 
		rules_one_hot = tf.one_hot(rules_int, self.num_rules_to_train)
		rules_int = tf.expand_dims(rules_int, axis=-1)
		w_dict = {'x': x, 'rules' : rules_one_hot,
				'rules_int': rules_int}
		w_logits = self.w_network(w_dict, dropout_keep_prob=self.dropout_keep_prob)
		w_logits = tf.squeeze(w_logits)
		weights = tf.nn.sigmoid(w_logits)
		return weights, w_logits

	def joint_scores_from_f_and_w(self,weights,m,f_probs):
		'''
		Func desc:
		Compute the learning scores obtained while jointly learning f(classification network) and w(rule network)

		Input:
		self object
		weights([num_instances, num_rules]) - the weights matrix corresponding to rule network(w_network) in the algorithm
		m([batch_size, num_rules]) - the rule coverage matrix where m_ij denotes if jth rule covers ith instance (if yes, then m_ij = 1)
		f_probs([batch_size, 1]) - the prob values from classification network (f_network)

		Output:
		results([batch_size,1]) -  
		'''
		num_classes = self.num_classes
		rule_classes = self.rule_classes
		#[batch_size, num_rules, 1]
		weights = tf.expand_dims(weights,-1)
		weights_mask = tf.to_float(tf.greater(weights,0.5))
		#[batch_size, num_rules, 1]
		m = tf.expand_dims(m,-1)
		m = m*weights_mask
		#[num_rules, num_classes]
		one_hot_rule_classes = tf.one_hot(rule_classes,num_classes,dtype=tf.float32)
		#[1, num_rules, num_classes]
		one_hot_rule_classes = tf.expand_dims(one_hot_rule_classes,0)
		#[batch_size, num_rules, num_classes]
		rule_weight_product = weights * one_hot_rule_classes + (1-weights)*(1-one_hot_rule_classes)
		sum_rule_firings = tf.reduce_sum(m,1)
		result = m*rule_weight_product #+ (1-m)
		#[batch_size, num_classes]
		result = tf.reduce_sum(result,1)/(sum_rule_firings+1e-20)
		result = result + f_probs
		return result    



import os
import glob
import json
import numpy as np
# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()


class BestCheckpointSaver(object):
    '''Maintains a directory containing only the best n checkpoints

    Inside the directory is a best_checkpoints JSON file containing a dictionary
    mapping of the best checkpoint filepaths to the values by which the checkpoints
    are compared.  Only the best n checkpoints are contained in the directory and JSON file.

    This is a light-weight wrapper class only intended to work in simple,
    non-distributed settings.  It is not intended to work with the tf.Estimator
    framework.
    '''
    def __init__(self, save_dir, num_to_keep=1, maximize=True, saver=None):
        '''
        Func Desc:
        Creates a `BestCheckpointSaver`
        `BestCheckpointSaver` acts as a wrapper class around a `tf.train.Saver`

        Input:
        save_dir: The directory in which the checkpoint files will be saved
        num_to_keep: The number of best checkpoint files to retain
        maximize: Define 'best' values to be the highest values.  For example,
            set this to True if selecting for the checkpoints with the highest
            given accuracy.  Or set to False to select for checkpoints with the
            lowest given error rate.
        saver: A `tf.train.Saver` to use for saving checkpoints.  A default
            `tf.train.Saver` will be created if none is provided.

        Output:

        '''
        self._num_to_keep = num_to_keep
        self._save_dir = save_dir
        self._save_path = os.path.join(save_dir, 'best.ckpt')
        self._maximize = maximize
        self._saver = saver if saver else tf.train.Saver(
            max_to_keep=None,
            save_relative_paths=True
        )

        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        self.best_checkpoints_file = os.path.join(save_dir, 'best_checkpoints')

    def handle(self, value, sess, global_step_tensor):
        '''
        Func Desc:
        Updates the set of best checkpoints based on the given result.

        Input:
        value: The value by which to rank the checkpoint.
        sess: A tf.Session to use to save the checkpoint
        global_step_tensor: A `tf.Tensor` represent the global step

        Output:
        True or False
        '''
        global_step = sess.run(global_step_tensor)
        current_ckpt = 'best.ckpt-{}'.format(global_step)
        value = float(value)
        if not os.path.exists(self.best_checkpoints_file):
            self._save_best_checkpoints_file({current_ckpt: value})
            self._saver.save(sess, self._save_path, global_step_tensor)
            return True

        best_checkpoints = self._load_best_checkpoints_file()

        if len(best_checkpoints) < self._num_to_keep:
            best_checkpoints[current_ckpt] = value
            self._save_best_checkpoints_file(best_checkpoints)
            self._saver.save(sess, self._save_path, global_step_tensor)
            return True

        if self._maximize:
            should_save = not all(current_best >= value
                                  for current_best in best_checkpoints.values())
        else:
            should_save = not all(current_best <= value
                                  for current_best in best_checkpoints.values())
        if should_save:
            best_checkpoint_list = self._sort(best_checkpoints)

            worst_checkpoint = os.path.join(self._save_dir,
                                            best_checkpoint_list.pop(-1)[0])
            self._remove_outdated_checkpoint_files(worst_checkpoint)
            self._update_internal_saver_state(best_checkpoint_list)

            best_checkpoints = dict(best_checkpoint_list)
            best_checkpoints[current_ckpt] = value
            self._save_best_checkpoints_file(best_checkpoints)

            self._saver.save(sess, self._save_path, global_step_tensor)
            return True
        return False

    def _save_best_checkpoints_file(self, updated_best_checkpoints):
        '''
        Func Desc:
        Save the best checkpoints

        Input:
        self 
        updated_best_checkpoints

        Output:

        '''
        with open(self.best_checkpoints_file, 'w') as f:
            json.dump(updated_best_checkpoints, f, indent=3)

    def _remove_outdated_checkpoint_files(self, worst_checkpoint):
        '''
        Func Desc:
        To remove the outdated checkpoint files

        Input:
        self 
        worst_checkpoint

        Output:

        '''
        os.remove(os.path.join(self._save_dir, 'checkpoint'))
        for ckpt_file in glob.glob(worst_checkpoint + '.*'):
            os.remove(ckpt_file)

    def _update_internal_saver_state(self, best_checkpoint_list):
        '''
        Func Desc:
        TO update the internal saver state with best_checkpoint_list

        Input:
        self
        best_checkpoint_list

        Output:

        '''
        best_checkpoint_files = [
            (ckpt[0], np.inf)  # TODO: Try to use actual file timestamp
            for ckpt in best_checkpoint_list
        ]
        self._saver.set_last_checkpoints_with_time(best_checkpoint_files)

    def _load_best_checkpoints_file(self):
        '''
        Func Desc:
        load rhe best checkpoints file

        Input:
        self

        Output:
        the best checkpoints
        '''
        with open(self.best_checkpoints_file, 'r') as f:
            best_checkpoints = json.load(f)
        return best_checkpoints

    def _sort(self, best_checkpoints):
        '''
        Func Desc: 
        Sort the best_checkpoints list in the decreasing order of their goodness

        Input:
        self
        best_checkpoints - list of checkpoints

        Output:
        
        '''
        best_checkpoints = [
            (ckpt, best_checkpoints[ckpt])
            for ckpt in sorted(best_checkpoints,
                               key=best_checkpoints.get,
                               reverse=self._maximize)
        ]
        return best_checkpoints


def get_best_checkpoint(best_checkpoint_dir, select_maximum_value=True):
    ''' 
    Func Desc:
    Reads the best_checkpoints file in the best_checkpoint_dir directory.
    Returns the filepath in the best_checkpoints file associated with
    the highest value if select_maximum_value is True, or the filepath
    associated with the lowest value if select_maximum_value is False.

    Input:
    best_checkpoint_dir: Directory containing best_checkpoints JSON file
    select_maximum_value: If True, select the filepath associated
    with the highest value.  Otherwise, select the filepath associated
    with the lowest value.

    Output:
    The full path to the best checkpoint file

    '''
    best_checkpoints_file = os.path.join(best_checkpoint_dir, 'best_checkpoints')
    if not os.path.exists(best_checkpoints_file):
        raise ValueError('Checkpoint file does not exist')

    with open(best_checkpoints_file, 'r') as f:
        best_checkpoints = json.load(f)
    best_checkpoints = [
        ckpt for ckpt in sorted(best_checkpoints,
                                key=best_checkpoints.get,
                                reverse=select_maximum_value)
    ]
    return os.path.join(best_checkpoint_dir,  best_checkpoints[0])



import enum
import numpy as np
from tqdm import tqdm
from itertools import chain
from typing import DefaultDict, Dict, List, Set, NamedTuple, Tuple, Union

from ..lf_set import LFSet
from ..utils import check_unique_names
from ..lf import ABSTAIN, LabelingFunction
from ..data_types import DataPoint, DataPoints


RowData = List[Tuple[int, int, int, float]]     # index of datapoint, index of lf, label, confidence


class ApplierMetadata(NamedTuple):
    """Metadata about Applier call."""

    # Map from LF name to number of faults in apply call
    faults: Dict[str, int]


class _FunctionCaller:
    def __init__(self, fault_tolerant: bool):
        self.fault_tolerant = fault_tolerant
        self.fault_counts: DefaultDict[str, int] = DefaultDict(int)

    def __call__(self, f: LabelingFunction, x: DataPoint):
        if not self.fault_tolerant:
            return f(x)
        try:
            return f(x)
        except Exception:
            self.fault_counts[f.name] += 1
            return -1


class BaseLFApplier:
    """Base class for LF applier objects.
    Base class for LF applier objects, which executes a set of LFs
    on a collection of data points. Subclasses should operate on
    a single data point collection format (e.g. ``DataFrame``).
    Subclasses must implement the ``apply`` method.

    Args:
        lf_set (LFSet): Instace of LFset which has information of set of labeling functions(which is applied on data)
    
    Raises:
        ValueError:
            If names of LFs are not unique
    """
    _use_recarray = False

    def __init__(self, lf_set: LFSet) -> None:    
        # self._lf_set = lf_set
        self._lfs = lf_set.get_lfs()
        self._lf_names = [lf.name for lf in lf_set.get_lfs()]
        # self._enum = enum
        check_unique_names(self._lf_names)

    def _numpy_from_row_data(self, labels: List[RowData]) -> np.ndarray:
        L = np.empty((len(labels), len(self._lfs)), dtype=object)
        L.fill(ABSTAIN)
        S = np.full((len(labels), len(self._lfs)), None) #np.zeros((len(labels), len(self._lfs)), dtype=float) - 1.0
        # NB: this check will short-circuit, so ok for large L
        if any(map(len, labels)):
            row, col, enm, conf = zip(*chain.from_iterable(labels))
            L[row, col] = enm
            S[row, col] = conf

        if self._use_recarray:                                        # always false
            n_rows, _ = L.shape
            dtype = [(name, np.int64) for name in self._lf_names]
            recarray = np.recarray(n_rows, dtype=dtype)
            for idx, name in enumerate(self._lf_names):
                recarray[name] = L[:, idx]

            return recarray
        else:
            return L,S

    def __repr__(self) -> str:
        return f"{type(self).__name__}, LFs: {self._lf_names}"


def apply_lfs_to_data_point(
    x: DataPoint, index: int, lfs: List[LabelingFunction], f_caller: _FunctionCaller
) -> RowData:
    """Label a single data point with a set of LFs

    Args:
        x (DataPoint): Data point to label
        index (int): Index of the data point
        lfs (List[LabelingFunction]): List of LFs to label ``x`` with
        f_caller (_FunctionCaller): A ``_FunctionCaller`` to record failed LF executions

    Returns:
        RowData: A list of (data point index, LF index, label enum, confidence) tuples
    """
    labels = []
    for j, lf in enumerate(lfs):
        y, z = f_caller(lf, x)
        if (y==ABSTAIN and z is None):
            continue
        if (y==ABSTAIN and z is not None):
            labels.append((index, j, y, z))
            continue    
        assert(lf._label == y)
        labels.append((index, j, y.value, z))
    return labels


class LFApplier(BaseLFApplier):
    """LF applier for a list of data points (e.g. ``SimpleNamespace``) or a NumPy array.

    Args:
        lf_set (LFSet): Instace of LFset which has information of set of labeling functions(which is applied on data)
    """ 

    # Example:
    #     >>> from labeling.lf import labeling_function
    #     >>> @labeling_function()
    #     ... def is_big_num(x,**kwargs):
    #     ...     return 1 if x.num > 42 else 0
    #     >>> applier = LFApplier([is_big_num])
    #     >>> from types import SimpleNamespace
    #     >>> applier.apply([SimpleNamespace(num=10), SimpleNamespace(num=100)])
    #     array([[0], [1]])
    #     >>> @labeling_function()
    #     ... def is_big_num_np(x):
    #     ...     return 1 if x[0] > 42 else 0
    #     >>> applier = LFApplier([is_big_num_np])
    #     >>> applier.apply(np.array([[10], [100]]))
    #     array([[0], [1]])  

    def apply(
        self,
        data_points: Union[DataPoints, np.ndarray],
        progress_bar: bool = True,
        fault_tolerant: bool = False,
        return_meta: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, ApplierMetadata]]:
        """Label list of data points or a NumPy array with LFs.

        Args:
            data_points (Union[DataPoints, np.ndarray]): List of data points or NumPy array to be labeled by LFs
            progress_bar (bool, optional): Display a progress bar?. Defaults to True.
            fault_tolerant (bool, optional): Output ``-1`` if LF execution fails?. Defaults to False.
            return_meta (bool, optional): Return metadata from apply call?. Defaults to False.

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, ApplierMetadata]]:
                np.ndarray:
                    Matrix of labels emitted by LFs
                ApplierMetadata:
                    Metadata, such as fault counts, for the apply call
        """        
        labels = []
        f_caller = _FunctionCaller(fault_tolerant)
        if progress_bar:
            with tqdm(total=len(data_points)) as pbar:
                for i, x in enumerate(data_points):
                    labels.append(apply_lfs_to_data_point(x, i, self._lfs, f_caller))
                    pbar.update()
        else:
            for i, x in enumerate(data_points):
                labels.append(apply_lfs_to_data_point(x, i, self._lfs, f_caller))
        
        L,S = self._numpy_from_row_data(labels)
        if return_meta:
            return L, ApplierMetadata(f_caller.fault_counts)
        return L,S


from .lf_set import LFSet
from .analysis import LFAnalysis
from .preprocess import preprocessor
from .prelabels import PreLabels
from .lf import labeling_function, ABSTAIN
from .continuous_scoring import continuous_scorer


# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np
import os

# from .my_checkmate import BestCheckpointSaver, get_best_checkpoint
# from .my_data_types import train_modes
from .checkmate import BestCheckpointSaver, get_best_checkpoint
from .data_types import train_modes
checkpoint_dir = "./checkpoint"
if not os.path.exists(checkpoint_dir):
	os.makedirs(checkpoint_dir)

num_checkpoints = 1 # Number of checkpoints to keep around

# Keeps only the most recently saved checkpoint
#
# max_to_keep is deliberatly set to 1 in order to provide for the case when more recent checkpoint
# has a smaller global_step. tf.train.Saver() orders by global_step.

class MRUCheckpoint():
	def __init__(self, path, session, variables):
		'''
		Func Desc:
		Initializes the class variables

		Input:
		self
		path - file path
		session 
		variables

		Output:

		'''
		self.ckpt_path = path
		self.ckpt_file = os.path.join(path, 'checkpoint')
		self.checkpoint_prefix = os.path.join(self.ckpt_path, 'hls-model')
		self.sess = session
		# max_to_keep
		self.saver = tf.train.Saver(variables, max_to_keep=1)
		# self.saver = tf.train.Saver()

	def save(self, global_step=None):
		'''
		Func Desc:
		saves the obtained checkpoint

		Input:
		self
		global step (Default - none)

		Output:

		'''
		path = self.saver.save(self.sess, self.checkpoint_prefix, global_step)
		print('Saved MRU checkpoint to path: ', path)
		
	def restore(self):
		'''
		Func Desc:
		Restores the last checkpoint

		Input:
		self

		Output:

		'''
		last_checkpoint = tf.train.latest_checkpoint(self.ckpt_path, 'checkpoint')
		#if self.saver.last_checkpoints:
		#    last_checkpoint = self.saver.last_checkpoints[0]
		#    print('All saved checkpoints: ', self.saver.last_checkpoints)
		#else:
		if not last_checkpoint:
			last_checkpoint = self.checkpoint_prefix

		print('Restoring checkpoint from path: ', last_checkpoint)
		self.saver.restore(self.sess, last_checkpoint)

	def restore_if_checkpoint_exists(self):
		'''
		Func Desc:
		checks if there exists any checkpoint for the file 

		Input:
		self

		Output:
		Boolean (True or False)
		'''
		if os.path.exists(self.ckpt_file):
			self.restore()
			return True
		return False

def test_mru_checkpoints(num_to_keep):
	'''
	Func Desc:
	Runs different sessions while changing the checkpoint number that is currently being worked with and tests the same
	
	Input:
	num_to_keep(int) - a limit on the size of the global step for checkpoint traversal

	Output:

	'''
	global_step = tf.get_variable(name='mru_global_step_%d' % num_to_keep, initializer=10, dtype=tf.int32)
	inc = tf.assign_add(global_step, 1)
	sess = tf.Session()
	sess.run(tf.global_variables_initializer())
	assert sess.run(global_step) == 10

	sess.run(inc)
	assert sess.run(global_step) == 11
	ckpt_path = '/tmp/checkpoints_%.6f' % np.random.rand()
	ckpt = MRUCheckpoint(ckpt_path, sess, tf.global_variables())
	ckpt.save(global_step)

	sess.run(inc)
	assert sess.run(global_step) == 12

	ckpt.restore_if_checkpoint_exists()
	assert sess.run(global_step) == 11

	assgn_op = tf.assign(global_step, 5)
	sess.run(assgn_op)
	assert sess.run(global_step) == 5
	ckpt.save(global_step)

	sess.run(inc)
	sess.run(inc)
	assert sess.run(global_step) == 7

	ckpt.restore_if_checkpoint_exists()
	assert sess.run(global_step) == 5

def test_checkpoint():
	'''
	Func Desc:
	tests whether the checkpoints stored are as expected

	Input:

	Output:

	'''
	v = tf.get_variable(name='v', initializer=12, dtype=tf.int32)
	v1 = tf.assign_add(v, 1)
	sess = tf.Session()
	sess.run(tf.global_variables_initializer())
	assert sess.run(v) == 12
	sess.run(v1)
	assert sess.run(v) == 13
	ckpt = MRUCheckpoint('/tmp/checkpoints', sess, tf.global_variables())
	ckpt.save()

	sess1 = tf.Session()
	sess1.run(tf.global_variables_initializer())
	assert sess1.run(v) == 12
	ckpt1 = MRUCheckpoint('/tmp/checkpoints', sess1, tf.global_variables())
	ckpt1.restore()
	assert sess1.run(v) == 13
	sess1.run(v1)
	assert sess1.run(v) == 14

	ckpt2 = MRUCheckpoint('/tmp/bad-ckpt-path', sess1, tf.global_variables())
	restored = ckpt2.restore_if_checkpoint_exists()
	assert restored == False

	restored = ckpt1.restore_if_checkpoint_exists()
	assert restored
	assert sess1.run(v) == 13

class BestCheckpoint():
	def __init__(self, path, prefix, session, num_checkpoints, variables, global_step):
		'''
		Func Desc:
		Initializes the class member variables to find the Best checkpoint so far

		Input:
		self
		path
		prefix
		session
		num_checkpoints
		variables
		global_step

		Output:

		'''
		self.ckpt_path = os.path.join(path, prefix)
		#self.ckpt_file = os.path.join(self.ckpt_path, 'checkpoint')
		#self.checkpoint_prefix = os.path.join(self.ckpt_path, prefix)
		self.sess = session
		# max_to_keep is None. Number of checkpoints is handled separately by BestCheckpointSaver
		self.saver = tf.train.Saver(variables, max_to_keep=None, save_relative_paths=True) 
		# self.saver = tf.train.Saver()
		self.best_ckpt_saver = BestCheckpointSaver(
			save_dir=self.ckpt_path,
			num_to_keep=num_checkpoints,
			maximize=True,
			saver=self.saver
			)
		self.global_step = global_step

	def save_if_best(self, metric):
		'''
		Func Desc:
		save if the current checkpoint is the best so far

		Input:
		self
		metric

		Output:

		'''
		saved = self.best_ckpt_saver.handle(metric, self.sess, self.global_step)
		path = tf.train.latest_checkpoint(self.ckpt_path, 'checkpoint')
		if saved:
			print('Saved new best checkpoint to path: ', path)
		else:
			print('No new best checkpoint. Did not save a new best checkpoint. Last checkpointed file: ', path)
		
	def restore_best_checkpoint(self):
		'''
		Func Desc:
		Restore the best checkpoint so far

		Input:
		self

		Output:

		'''
		best_ckpt_file = get_best_checkpoint(self.ckpt_path, select_maximum_value=True)
		print('Restoring best checkpoint from path: ', best_ckpt_file)
		self.saver.restore(self.sess, best_ckpt_file)

	def restore_best_checkpoint_if_exists(self):
		'''
		Func Desc:
		Restore the best checkpoint so far only if it exists

		Input:
		self

		Output:

		'''
		try:
			self.restore_best_checkpoint()
			return True
		except ValueError as e:
			print(str(e))
			return False

def test_best_ckpt():
	'''
	Func Desc:
	test for the best checkpoint so far

	Input:

	Output:

	'''
	global_step = tf.get_variable(name='global_step', initializer=50, dtype=tf.int32)
	inc_global_step = tf.assign_add(global_step, 1)
	sess1 = tf.Session()
	sess2 = tf.Session()
	
	sess1.run(tf.global_variables_initializer())
	sess2.run(tf.global_variables_initializer())

	# We'll save using sess1 and restore in sess2
	best_checkpoint_dir = '/tmp/best_ckpt_%.6f' % np.random.rand()
	best1 = BestCheckpoint(best_checkpoint_dir, 'foo-bar', sess1, 3, tf.trainable_variables(), global_step)
	best2 = BestCheckpoint(best_checkpoint_dir, 'foo-bar', sess2, 3, tf.trainable_variables(), global_step)

	restored = best2.restore_best_checkpoint_if_exists()
	assert not restored

	sess1.run(inc_global_step) ## 51
	best1.save_if_best(0.1)

	assert sess2.run(global_step) == 50
	restored = best2.restore_best_checkpoint_if_exists()
	assert restored
	assert sess2.run(global_step) == 51

	sess1.run(inc_global_step) ## 52
	best1.save_if_best(0.05)

	sess2.run(inc_global_step) # 52
	sess2.run(inc_global_step) # 53
	sess2.run(inc_global_step) # 54
	assert sess2.run(global_step) == 54
	restored = best2.restore_best_checkpoint_if_exists()
	assert restored
	assert sess2.run(global_step) == 51

	sess1.run(inc_global_step) ## 53
	best1.save_if_best(0.2)
	sess1.run(inc_global_step) ## 54
	best1.save_if_best(0.15)

	sess2.run(inc_global_step) # 52
	sess2.run(inc_global_step) # 53
	sess2.run(inc_global_step) # 54
	sess2.run(inc_global_step) # 55
	assert sess2.run(global_step) == 55
	restored = best2.restore_best_checkpoint_if_exists()
	assert restored
	assert sess2.run(global_step) == 53

def test_checkmate():
	'''
	Func Desc:
	test whether the checkmate model is working fine

	Input:

	Output:

	'''
	global_step = tf.get_variable(name='checkmate_global_step', initializer=12, dtype=tf.int32)
	inc_global_step_op = tf.assign_add(global_step, 1)
	sess = tf.Session()
	sess1 = tf.Session()
	
	sess.run(tf.global_variables_initializer())
	saver = tf.train.Saver(tf.trainable_variables(), max_to_keep=None)
	# saver = tf.train.Saver()
	best_checkpoint_dir = '/tmp/best_ckpt_%.6f' % np.random.rand()
	best_ckpt_saver = BestCheckpointSaver(
			save_dir=best_checkpoint_dir,
			num_to_keep=3,
			maximize=True,
			saver=saver
			)

	accuracy = 0.1 # 12
	best_ckpt_saver.handle(accuracy, sess, global_step)

	accuracy = 0.2
	sess.run(inc_global_step_op) # 13
	best_ckpt_saver.handle(accuracy, sess, global_step)

	accuracy = 0.05
	sess.run(inc_global_step_op) # 14
	best_ckpt_saver.handle(accuracy, sess, global_step)

	ckpt_path = get_best_checkpoint(best_checkpoint_dir, select_maximum_value=True)
	print('Best ckpt path: ', ckpt_path)
	saver.restore(sess1, ckpt_path)
	assert sess1.run(global_step) == 13
	
	accuracy = 0.12
	sess.run(inc_global_step_op) # 15
	best_ckpt_saver.handle(accuracy, sess, global_step)

	ckpt_path = get_best_checkpoint(best_checkpoint_dir, select_maximum_value=True)
	print('Best ckpt path: ', ckpt_path)
	saver.restore(sess1, ckpt_path)
	assert sess1.run(global_step) == 13

	accuracy = 0.45
	sess.run(inc_global_step_op) # 16
	best_ckpt_saver.handle(accuracy, sess, global_step)

	ckpt_path = get_best_checkpoint(best_checkpoint_dir, select_maximum_value=True)
	print('Best ckpt path: ', ckpt_path)
	saver.restore(sess1, ckpt_path)
	assert sess1.run(global_step) == 16

	# Now select lowest value
	ckpt_path = get_best_checkpoint(best_checkpoint_dir, select_maximum_value=False)
	print('Best ckpt path: ', ckpt_path)
	saver.restore(sess1, ckpt_path)
	assert sess1.run(global_step) == 15

# Loading of checkpoints happens only once - at the end of HLSModel initialization.
#
# Saving of checkpoints happens during training. We have only one MRU checkpoint saver 
# We have one best checkpoint saver per train mode type
class CheckpointsFactory:
	def __init__(self, sess, global_steps):
		'''
		Func Desc:
		Initializes the class with the arguments

		Input:
		self
		sess 
		global_steps

		Output:

		'''
		self.best_savers = {}
		self.initialize_savers(sess, global_steps)

	def get_best_saver(self, train_mode):
		'''
		Func Desc:
		get the best saved checkpoints

		Input:
		self
		Train_mode - the mode of training

		Output:

		'''
		return self.best_savers[train_mode]

	def initialize_savers(self, sess, global_steps):
		'''
		Func Desc:
		Initialize the required savers

		Input:
		self
		sess
		global_steps

		Output:

		'''
		for mode in train_modes:
			self.init_saver(sess, mode, global_steps)

	def init_saver(self, sess, mode, global_steps):
		'''
		Func Desc:
		Initialize the required savers with the given mode

		Input:
		self
		sess
		mode
		global_steps

		Output:

		'''
		ckpt_dir = checkpoint_dir
		self.best_savers[mode] = BestCheckpoint(ckpt_dir, mode, sess,
				num_checkpoints, tf.global_variables(), global_steps[mode])




# used to feed data while training models

import numpy as np
import os
import pickle
import inspect
import time

from spear.Implyloss import *

# from .my_data_types import *
# from .my_data_feeder_utils import *
from .data_types import *
from .data_feeder_utils import *

class DataFeeder():
    def __init__(self, d_pickle, U_pickle, validation_pickle, map_json, 
            shuffle_batches, num_load_d, num_load_U, num_classes, 
            f_d_class_sampling, min_rule_coverage, rule_classes, num_load_validation, 
            f_d_batch_size, f_d_U_batch_size, test_w_batch_size, out_dir='./'):
        '''
        Func Desc:
        Initialize the object with the given parameter files

        Input: 
        self
        d_pickle - labelled data file
        U_pickle - unlabelled data file
        validation_pickle - validation data file
        out_dir (Default = './') - output directory

        Output:
        Void
        '''
        self.f_d_U_start = 0
        self.shuffle_batches=shuffle_batches
        self.out_dir = out_dir
        
        self.raw_d = load_data(d_pickle, map_json, num_load_d)
        self.raw_U = load_data(U_pickle, map_json, num_load_U)

        if num_classes is not None:
            self.num_classes = num_classes
            assert self.num_classes >= np.max(self.raw_d.L) + 1
        else:
            self.num_classes = np.max(self.raw_d.L) + 1

        self.f_d_class_sampling = f_d_class_sampling
        if self.f_d_class_sampling:
            assert len(self.f_d_class_sampling) == self.num_classes
        else:
            self.f_d_class_sampling = [1] * self.num_classes
        # Set f_d_U class sampling dist to be the same as f_d class sampling dist
        self.f_d_U_class_sampling = self.f_d_class_sampling

        self.phi = self.num_classes
        self.num_features = self.raw_d.x.shape[1]
        self.num_rules = self.raw_d.l.shape[1]

        # If min coverage threshold is specified for rules then apply it
        self.min_rule_coverage = min_rule_coverage
        self.num_rules_to_train = self.num_rules
        if self.min_rule_coverage > 0:
            self.satisfying_rules, self.not_satisfying_rules, \
                    self.rule_map_new_to_old, self.rule_map_old_to_new  = \
                    utils.extract_rules_satisfying_min_coverage(self.raw_U.m,
                            self.min_rule_coverage)
            self.num_rules_to_train = len(self.satisfying_rules)
            assert np.all(self.satisfying_rules ==
                    self.rule_map_new_to_old[0:self.num_rules_to_train])
            assert np.all(self.not_satisfying_rules ==
                    self.rule_map_new_to_old[self.num_rules_to_train:])
            print('Originally %d rules. To train on %d rules' %
                    (self.num_rules, self.num_rules_to_train))
            print('Rule map new to old: ', self.rule_map_new_to_old)
            if self.num_rules != self.num_rules_to_train:
                utils.modify_d_or_U_using_rule_map(self.raw_U,
                        self.rule_map_old_to_new)
                utils.modify_d_or_U_using_rule_map(self.raw_d,
                        self.rule_map_old_to_new)

        # Determine rule classes from the truncated rule list
        self.rule_classes = rule_classes 
        if not self.rule_classes:
            self.rule_classes = get_rule_classes(self.raw_d.l, self.num_classes)

        print('Rule classes: ', self.rule_classes)

        # Remove data from U for which no rule makes any predictions
        self.covered_U = self.remove_instances_labeled_by_no_rules(self.raw_U)
        print("length of covered U: {}".format(len(self.covered_U.x)))

        # Now combine d and U data
        self.f_d_U = self.combine_f_d_U(self.raw_d, self.covered_U, self.f_d_U_class_sampling)

        self.f_d = self.convert_raw_d_to_f_d(self.raw_d, num_load=0)

        raw_test_data = load_data(validation_pickle, map_json,
                num_load_validation)

        self.test_f_x, self.test_f_labels, self.test_f_labels_one_hot, \
                self.test_f_l, self.test_f_m, self.test_f_d, self.test_f_r  = \
                self.convert_raw_test_data_to_f(raw_test_data)

        self.test_w = self.convert_raw_test_data_to_w(raw_test_data)
        print('test_w len: ', len(self.test_w.x))

        self.batch_counter = {
                f_d:0,
                f_d_U:0,
                test_w:0,
                }

        self.data_lens = {
                f_d: len(self.f_d.x),
                f_d_U: len(self.f_d_U.x),
                test_w:len(self.test_w.x),
                }

        print('test_w len: ', self.data_lens[test_w])

        self.batch_size = {
                f_d: f_d_batch_size ,
                f_d_U: f_d_U_batch_size,
                test_w: test_w_batch_size,
                }

        self.data_store = {
                test_w: self.test_w,
                }

        self.shuf_indices = {}
        for data_type in [f_d, f_d_U]:
            self.shuf_indices[data_type] = np.arange(self.data_lens[data_type])
            self.reset_batch(data_type)

    
    def convert_raw_test_data_to_f(self, raw_test_data):
        '''
        Func Desc:
        to convert raw test data to f (classification network)

        Input:
        self
        raw_test_data - 

        Output:
        f data with the required parameters
        '''
        x = raw_test_data.x
        labels = np.squeeze(raw_test_data.L)
        assert max(labels) <= self.num_classes - 1 or np.all(labels == self.num_classes)
        labels_one_hot = np.eye(self.num_classes + 1)[labels][:, : -1]
        return x, labels, labels_one_hot, raw_test_data.l, raw_test_data.m, raw_test_data.d, raw_test_data.r

    def convert_raw_test_data_to_w(self, raw_test_data):
        '''
        Func Desc:
        to convert raw test data to w (rule network)

        Input:
        self
        raw_test_data - 

        Output:
        F_d_U_data
        '''
        test_w = self.remove_instances_labeled_by_no_rules(raw_test_data)
        print('Setting value of d to 0 for test data')
        d_new = np.zeros_like(test_w.d)
            
        return F_d_U_Data(test_w.x,
                test_w.l,
                test_w.m,
                test_w.L,
                d_new,
                test_w.r)


    def remove_instances_labeled_by_no_rules(self, raw_U):
        '''
        Func Desc:
        Removes those instances that are labelled by no rules

        Input:
        self
        raw_U - raw Unlabelled Data

        Output:
        F_d_U_data
        '''
        xx = []
        ll = []
        mm = []
        LL = []
        dd = []
        rr = []
        for x, l, m, L, d, r in zip(raw_U.x, raw_U.l, raw_U.m, raw_U.L, raw_U.d, raw_U.r):
            if np.all(l == self.phi):
                continue
            xx.append(x)
            ll.append(l)
            mm.append(m)
            LL.append(L)
            dd.append(d)
            rr.append(r)

        assert len(xx) == len(ll)
        assert len(xx) == len(mm)
        assert len(xx) == len(LL)
        assert len(xx) == len(dd)
        assert len(xx) == len(rr)
        return F_d_U_Data(np.array(xx),
            np.array(ll),
            np.array(mm),
            np.array(LL),
            np.array(dd),
            np.array(rr))


    def combine_f_d_U(self, raw_d, raw_U, d_class_sampling):
        '''
        Func Desc:
        combines the labelled (raw_d) and Unlabelled (raw_U) data

        Input:
        self
        raw_d - labelled data
        raw_U - unlabelled data
        d_class_sampling - sampling distribution
        
        '''
        print('Size of d before oversampling: ', len(raw_d.x))
        print('Size of U (covered) : ', len(raw_U.x))
        # Oversample d according to its class
        #
        # Note that we cannot oversample U according to their true labels since these should not be available
        # during training.
        raw_d = oversample_d(raw_d, d_class_sampling)
        print('Size of d after oversampling: ', len(raw_d.x))

        new_d_d = np.ones_like(raw_d.d)
        new_U_d = np.zeros_like(raw_U.d)

        #num_classes = np.max(raw_d.L) + 1
        #new_U_L = np.zeros_like(raw_U.L) + num_classes
        # We let the true U labels flow through for observation during f_d_U training
        new_U_L = raw_U.L


        xx = np.concatenate((raw_d.x, raw_U.x))
        ll = np.concatenate((raw_d.l, raw_U.l))
        mm = np.concatenate((raw_d.m, raw_U.m))
        LL = np.concatenate((raw_d.L, new_U_L))
        dd = np.concatenate((new_d_d, new_U_d))
        rr = np.concatenate((raw_d.r, raw_U.r))

        print('Size of d_U after combining: ', len(xx))
        f_d_U = F_d_U_Data(xx, ll, mm, LL, dd, rr)
        return f_d_U

    # Need x and true labels only (x, L)
    def convert_raw_d_to_f_d(self, raw_d, num_load=30):
        '''
        Func Desc:
        converts raw d to f

        Input:
        self
        raw_d - 
        num_load (default = 30)

        Output:
        F_d_Data
        '''
        if num_load <= 0:
            num_load = len(raw_d.x)

        print('Loading %d elements from d' % num_load)

        x = raw_d.x[0:num_load]
        label = np.squeeze(raw_d.L[0:num_load])
        x, label = oversample_f_d(x, label, self.f_d_class_sampling)
        print('num instances in d: ', len(x))
        return F_d_Data(x, label)    

    def reset_batch(self, data_type):
        '''
        Func Desc:
        restes the batch

        Input:
        self
        data_type

        Output:

        '''
        self.batch_counter[data_type] = 0
        if not self.shuffle_batches or data_type == test_w:
            print('Not shuffling batch for data type: ', data_type)
            return

        #print('Shuffling batch for data type: ', data_type)
        np.random.shuffle(self.shuf_indices[data_type])

        idx = self.shuf_indices[data_type]
        # We need to actually shuffle each array
        if data_type == f_d:
            np.take(self.f_d.x, idx, axis=0, out=self.f_d.x)
            np.take(self.f_d.labels, idx, axis=0, out=self.f_d.labels)

        elif data_type == f_d_U:
            np.take(self.f_d_U.x, idx, axis=0, out=self.f_d_U.x)
            np.take(self.f_d_U.l, idx, axis=0, out=self.f_d_U.l)
            np.take(self.f_d_U.m, idx, axis=0, out=self.f_d_U.m)
            np.take(self.f_d_U.L, idx, axis=0, out=self.f_d_U.L)
            np.take(self.f_d_U.d, idx, axis=0, out=self.f_d_U.d)
            np.take(self.f_d_U.r, idx, axis=0, out=self.f_d_U.r)
        else:
            raise ValueError('Data type not recognized: ', data_type)


    # Get next batch indices. Shuffle if necessary
    #
    # NOTE: We DO NOT skip the last (incomplete) batch
    def next_batch(self, data_type):
        '''
        Func Desc:
        get the next batch for computation

        Input:
        self
        data_type

        Output:
        start - start of the next batch
        end - end of the next batch

        '''
        batch_size = self.batch_size[data_type]

        num_instances = self.data_lens[data_type]
        total_batch = num_instances // batch_size
        remaining = num_instances % batch_size
        if remaining > 0 and (total_batch == 0 or 'test' in data_type):
            #print('Should not skip last batch')
            skip_last_batch = False
        else:
            #print('Should skip last batch')
            skip_last_batch = True

        if skip_last_batch:
            check = self.batch_counter[data_type] * batch_size + batch_size > self.data_lens[data_type]
        else:
            check = self.batch_counter[data_type] * batch_size >= self.data_lens[data_type]

        #print('check is: ', check)
        if check:
            #print('Resetting batch')
            self.reset_batch(data_type)

        start = self.batch_counter[data_type] * batch_size
        end = min(start + batch_size, self.data_lens[data_type])
        self.batch_counter[data_type] += 1
        return start, end


    # x: [batch_size, num_features]
    # y: [batch_size, num_classes] --> one-hot
    #
    # from d data
    def get_f_d_next_batch(self):
        '''
        Func Desc:
        get the next batch in f_d (labelled data)

        Input:
        self

        Output:
        x ([batch_size, num_features]) - the data
        labels_one_hot
        '''
        if True:
            start, end = self.next_batch(f_d)
        else:
            if np.random.rand() < 0.17:
                self.reset_batch(f_d)
            self.f_d_U_start = (self.f_d_U_start + 1) % 6
            start = self.f_d_U_start
            end = np.random.geometric(0.5)
            end = end + start
            end = min(end, 6)

        x = self.f_d.x[start:end]
        labels = self.f_d.labels[start:end]

        labels_one_hot = np.eye(self.num_classes)[labels]
        return x, labels_one_hot


    def get_f_d_U_next_batch(self):
        '''
        Func Desc:
        get the next batch in f_d_U (labelled + unlabelled data)

        Input:
        self

        Output:
        x ([batch_size, num_features]) - the data
        l ([batch_size, num_rules]) - the data labels
        m ([batch_size, num_rules]) - rule association matrix
        L ([batch_size, 1]) - labelling check vector
        d ([batch_size, 1]) - labelled data check vector
        r ([batch_size, num_rules]) - rule coverage matrix
        '''
        start, end = self.next_batch(f_d_U)

        x = self.f_d_U.x[start:end]
        l = self.f_d_U.l[start:end]
        m = self.f_d_U.m[start:end]
        L = self.f_d_U.L[start:end]
        d = self.f_d_U.d[start:end]
        r = self.f_d_U.r[start:end]
        return x, l, m, L, d, r

    # Number of instances
    def get_f_d_num_instances(self):
        '''
        Func Desc:
        gives the number of data instances in f_d

        Input:
        self

        Output:
        the required count

        '''
        return len(self.f_d.x)

    def get_f_d_U_num_instances(self):
        '''
        Func Desc:
        gives the number of data instances in f_d_U

        Input:
        self

        Output:
        the required count

        '''
        return len(self.f_d_U.x)

    # Batch Sizes
    def get_f_d_batch_size(self):
        '''
        Func Desc:
        gives the batch_size in f_d

        Input:
        self

        Output:
        the required size

        '''
        return self.batch_size[f_d]

    def get_f_d_U_batch_size(self):
        '''
        Func Desc:
        gives the batch_size in f_d_U

        Input:
        self

        Output:
        the required size

        '''
        return self.batch_size[f_d_U]

    def get_batch_size(self, data_type):
        '''
        Func Desc:
        gives the batch_size of the required data type

        Input:
        self
        dat_type

        Output:
        the required size

        '''
        return self.batch_size[data_type]

    def get_batches_per_epoch(self, data_type):
        '''
        Func Desc:
        gives the total number of batches in the required data type

        Input:
        self
        dat_type

        Output:
        the required count

        '''
        num_instances = self.data_lens[data_type]
        batch_size = self.batch_size[data_type]
        total_batch = num_instances // batch_size
        remaining = num_instances % batch_size
        print('num_instances: ', num_instances )
        print('batch_size: ', batch_size )
        print('total_batch: ', total_batch )
        print('remaining: ', remaining )
        # Add last batch if it is the only batch
        # Else last batch is discarded
        #
        # Unless we are in test mode and the entire dataset needs to be tested
        if remaining > 0 and (total_batch == 0 or 'test' in data_type):
            total_batch += 1

        print('total_batch: ', total_batch )
        #print('instances\t batch_size\t total_batch\t remaining')
        #print('%d\t %d\t %d\t %d' % (num_instances, batch_size, total_batch,
        #    remaining))
        return total_batch

    def get_features_classes_rules(self):
        '''
        Func Desc:
        get the features, classes and rules of the object

        Input:
        self

        Output:
        num_features
        num_classes
        num_rules
        num_rules_to_train

        '''
        return self.num_features, self.num_classes, self.num_rules, \
                self.num_rules_to_train

    def get_f_test_data(self, data_type):
        '''
        Func Desc:
        get the test data for f_network

        Input:
        self
        data_type

        Output:
        test_f_x
        test_f_labels_one_hot
        test_f_L
        test_f_m
        test_f_d

        '''
        return self.test_f_x, self.test_f_labels_one_hot, \
                self.test_f_l, self.test_f_m, self.test_f_d

    def get_w_test_data(self, data_type='test_w'):
        '''
        Func Desc:
        get the test data for w_network

        Input:
        self
        data_type (fixed to test_w)

        Output:
        x ([batch_size, num_features]) - the data
        l ([batch_size, num_rules]) - the data labels
        m ([batch_size, num_rules]) - rule association matrix
        L ([batch_size, 1]) - labelling check vector
        d ([batch_size, 1]) - labelled data check vector
        
        '''
        assert data_type  == test_w
        start, end = self.next_batch(data_type)

        x = self.data_store[data_type].x[start:end]
        l = self.data_store[data_type].l[start:end]
        m = self.data_store[data_type].m[start:end]
        L = self.data_store[data_type].L[start:end]
        d = self.data_store[data_type].d[start:end]

        return x, l, m, L, d




'''
The common utils to CAGE and JL algorithms are in this file. Don't change the name or location of this file.
'''

import pickle, json
import numpy as np 
from os import path as check_path

def is_dict_trivial(dict):
	'''	
		A helper function that checks if the dictionary have key and value equal values for all keys except if its null

	Args:
		dict: the dictionary
	
	Return:
		True if all keys(which are not None) are equal to respective values. False otherwise
	'''
	for key, value in dict.items():
		if not(key == None):
			try:
				if key == value:
					continue
				else:
					return False
			except:
				return False
	return True


def get_data(path, check_shapes = True, class_map = None):
	'''
		Standard format in pickle file contains the NUMPY ndarrays x, l, m, L, d, r, s, n, k and an int n_classes
			x: (num_instances, num_features), x[i][j] is jth feature of ith instance. Note that the dimension fo this array can vary depending on the dimension of input
			
			l: (num_instances, num_lfs), l[i][j] is the prediction of jth LF(co-domain: the values used in Enum) on ith instance. l[i][j] = None imply Abstain
			
			m: (num_instances, num_lfs), m[i][j] is 1 if jth LF didn't Abstain on ith instance. Else it's 0
			
			L: (num_instances, 1), L[i] is true label(co-domain: the values used in Enum) of ith instance, if available. Else L[i] is None
			
			d: (num_instances, 1), d[i] is 1 if ith instance is labelled. Else it is 0
			
			r: (num_instances, num_lfs), r[i][j] is 1 if ith instance is an exemplar for jth rule. Else it's 0
			
			s: (num_instances, num_lfs), s[i][j] is the continuous score of ith instance given by jth continuous LF. If jth LF is not continuous, then s[i][j] is None
			
			n: (num_lfs,), n[i] is 1 if ith LF has continuous counter part, else n[i] is 0
			
			k: (num_lfs,), k[i] is the class of ith LF, co-domain: the values used in Enum

			n_classes: total number of classes

			In case the numpy array is not available(can be possible for x, L, d, r, s), it is stored as numpy.zeros(0)

	Args: 
		path: path to pickle file with data in the format above
		check_shapes: if true, checks whether the shapes of numpy arrays in pickle file are consistent as per the format mentioned above. Else it doesn't check. Default is True. 
		class_map: dictionary of class numbers(sorted, mapped to [0,n_classes-1]) are per the Enum defined in labeling part. l,L are modified(needed inside algorithms) before returning, using class_map. Default is None which doesn't do any mapping

	Return:
		A list containing all the numpy arrays mentioned above. The arrays l, L are modified using the class_map 
	'''
	assert check_path.exists(path)
	data = []
	with open(path, 'rb') as file:
		for i in range(9):
			data.append(pickle.load(file))
			assert type(data[i]) == np.ndarray
		data.append(pickle.load(file))
	
	assert type(data[9]) == np.int

	if check_shapes:
		assert data[1].shape == data[2].shape # l, m
		assert (data[1].shape == data[5].shape) or (data[5].shape[0] == 0) # l, r
		assert (data[1].shape == data[6].shape) or (data[6].shape[0] == 0) # l, s
		assert (data[3].shape == (data[1].shape[0],1)) or (data[3].shape[0] == 0) #L, l
		assert (data[4].shape == (data[1].shape[0],1)) or (data[4].shape[0] == 0) #d, l
		assert data[7].shape == (data[1].shape[1],) #n, l
		assert data[8].shape == (data[1].shape[1],) #k, l
		assert (data[0].shape[0] == 0) or data[1].shape[0] == data[0].shape[0] #x, l
		assert np.all(np.logical_or(data[2] == 0, data[2] == 1)) #m
		assert (data[4].shape[0] == 0) or (np.all(np.logical_or(data[4] == 0, data[4] == 1))) #d
		assert (data[5].shape[0] == 0) or (np.all(np.logical_or(data[5] == 0, data[5] == 1)) )#r
		assert np.all(np.logical_or(data[7] == 0, data[7] == 1)) #n

	if class_map == None:
		return data

	is_dict_trivial_ = is_dict_trivial(class_map)
	if not(is_dict_trivial_):
		data[1] = np.vectorize(class_map.get)(data[1])
		if not(data[3].shape[0] == 0):
			data[3] = np.vectorize(class_map.get)(data[3])
	else:
		data[1][data[1] == None] = data[9]
		if not(data[3].shape[0] == 0):
			data[3][data[3] == None] = data[9]

	data[6][data[6] == None] = 0 # s will have None values if LF is not continuous
	for i in range(9):
		if i == 0 or data[i].shape == 0:
			continue
		elif i == 6:
			data[i] = data[i].astype(np.float32)
		else:
			data[i] = data[i].astype(np.int32)

	return data

def get_classes(path):
	'''
		The json file should contain a dictionary of number to string(class name) map as defined in Enum

		Args:
			path: path to json file with contents mentioned above
		
		Returns:
			A dictionary (number to string(class name) map)
	'''
	assert check_path.exists(path)
	json_object = None
	with open(path, 'r') as f:
		json_object = json.load(f)
	json_object = {int(index): value for index, value in json_object.items()}
	return json_object

def get_predictions(proba, class_map, class_dict, need_strings):
	'''
		This function takes probaility of instances being a class and gives what class each instance belongs to, using the maximum of probabilities

	Args:
		proba: probability numpy.ndarray of shape (num_instances, num_classes)
		class_map: dictionary mapping the class numbers(as per Enum class defined) to numbers in range [0, num_classes-1]
		class_dict: dictionary consisting of number to string(class name) mapping as per the Enum class defined
		need_trings: If True, the output conatians strings(of class names), else it consists of numbers(class numbers as used in Enum definition)

	Return:
		numpy.ndarray of shape (num_instances,), where elements are class_names/class_numbers depending on need_strings is True/False, where the elements
		represent the class of each instance
	'''
	final_labels = np.argmax(proba, 1) # this is actually labels_with_altered_class_values
	if not(is_dict_trivial(class_map)):
		remap_dict = {value:index for index, value in (class_map).items()}
		final_labels = np.vectorize(remap_dict.get)(final_labels)
	if need_strings:
		class_dict_with_abstain = (class_dict).copy()
		class_dict_with_abstain[None] = 'ABSTAIN'
		return np.vectorize(class_dict_with_abstain.get)(final_labels)
	else:
		return final_labels

def get_enum(np_array, enm):
	'''
		This function is used to convert a numpy array of numbers to a numpy array of enums based on the Enum class provided 'enm'

	Args:
		np_array: a numpy.ndarray of any shape consisting of numbers
		enm: An class derived from 'Enum' class, which must contain map from every number in np_array to an enum

	Return:
		numpy.ndarray of shape shape as np_array but now contains enums(as per the mapping in 'enm') instead of numbers
	'''
	try:
		ans = np.vectorize(enm)(np_array)
	except:
		print("Error in get_enum function in utils.py: maybe enm doesn't containt the map for the numbers in np_array")
		exit(1)

	return ans



from submodlib.functions.facilityLocation import FacilityLocationFunction

import numpy as np
import torch
from sklearn.metrics.pairwise import euclidean_distances
import pickle
from os import path as check_path

from ..utils.data_editor import get_data, get_classes
from ..utils.utils_cage import  predict_gm_labels
from ..utils.utils_jl import find_indices, get_similarity_kernel

def rand_subset(n_all, n_instances):
	'''
		A function to choose random indices of the input instances to be labeled

	Args:
		n_all: number of available instances, type in integer
		n_intances: number of instances to be labelled, type is integer
	
	Return:
		A numpy.ndarray of the indices(of shape (n_sup,) and each element in the range [0,n_all-1)) to be labeled
	'''
	assert type(n_all) == np.int or type(n_all) == np.float
	assert type(n_instances) == np.int or type(n_instances) == np.float
	assert np.int(n_all) > np.int(n_instances)
	return np.sort(np.random.choice(int(n_all), int(n_instances), replace = False))

def unsup_subset(x_train, n_unsup):
	'''
		A function for unsupervised subset selection(the subset to be labeled)
		
	Args:
		x_train: A numpy.ndarray of shape (n_instances, n_features). All the data, intended to be used for training
		n_unsup: number of instances to be found during unsupervised subset selection, type is integer
	
	Return:
		numpy.ndarray of indices(shape is (n_sup,), each element lies in [0,x_train.shape[0])), the result of subset selection
	'''
	assert x_train.shape[0] > int(n_unsup)
	assert type(x_train) == np.ndarray
	
	use_cuda = torch.cuda.is_available()
	device = torch.device("cuda" if use_cuda else "cpu")
	torch.backends.cudnn.benchmark = True

	#fl = apricot.functions.facilityLocation.FacilityLocationSelection(random_state = 0, n_samples = int(n_unsup))
	#x_sub = fl.fit_transform(x_train)
	#indices = find_indices(torch.from_numpy(x_train).to(device=device), torch.from_numpy(np.array(x_sub)).to(device=device))

	fl = FacilityLocationFunction(n = x_train.shape[0], mode = "dense", data = x_train, metric = "euclidean")
	x_sub = fl.maximize(budget = int(n_unsup), optimizer = 'LazyGreedy', stopIfZeroGain = False, stopIfNegativeGain = False, verbose = False)
	indices = np.array([i[0] for i in x_sub])

	return np.sort(indices)

def sup_subset(path_json, path_pkl, n_sup, qc = 0.85):
	'''
		A helper function for supervised subset selection(the subset to be labeled) which just returns indices
	
	Args:
		path_json: Path to json file of number to string(class name) map
		path_pkl: Path to the pickle file containing all the training data in standard format
		n_sup: Number of instances to be found during supervised subset selection
		qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85

	Return:
		numpy.ndarray of indices(shape is (n_sup,), each element lies in [0,num_instances)), the result of subset selection AND the data which is list of contents of path_pkl
	'''
	assert (type(qc) == np.float and (qc >= 0 and qc <= 1)) or (type(qc) == np.ndarray and (np.all(np.logical_and(qc>=0, qc<=1)) ) )\
		 or (type(qc) == np.int and (qc == 0 or qc == 1))

	class_dict = get_classes(path_json)
	class_list = list((class_dict).keys())
	class_list.sort()
	n_classes = len(class_dict)

	class_map = {value : index for index, value in enumerate(class_list)}
	class_map[None] = n_classes

	use_cuda = torch.cuda.is_available()
	device = torch.device("cuda" if use_cuda else "cpu")
	torch.backends.cudnn.benchmark = True

	data = get_data(path_pkl, True, class_map)
	m = torch.abs(torch.tensor(data[2], device = device).long())
	s = torch.tensor(data[6], device = device).double() # continuous score
	assert m.shape[0] > int(n_sup)
	k = torch.tensor(data[8], device = device).long() # LF's classes
	n_lfs = m.shape[1]
	continuous_mask = torch.tensor(data[7], device = device).double() # Mask for s/continuous_mask
	qc_temp = torch.tensor(qc, device = device).double() if type(qc) == np.ndarray else qc
	params_1 = torch.ones((n_classes, n_lfs), device = device).double() # initialisation of gm parameters, refer section 3.4 in the JL paper
	params_2 = torch.ones((n_classes, n_lfs), device = device).double()

	y_train_pred = predict_gm_labels(params_1, params_2, m, s, k, n_classes, continuous_mask, qc_temp, device)
	kernel = get_similarity_kernel(y_train_pred)
	similarity = euclidean_distances(data[0])
	sim_mat = kernel * similarity

	#fl = apricot.functions.facilityLocation.FacilityLocationSelection(random_state = 0, metric = 'precomputed', n_samples = int(n_sup))
	#sim_sub = fl.fit_transform(sim_mat)
	#indices = find_indices(torch.from_numpy(sim_mat).to(device=device), torch.from_numpy(np.array(sim_sub)).to(device=device))

	fl = FacilityLocationFunction(n = sim_mat.shape[0], mode = "dense", sijs = sim_mat, separate_rep = False)
	sim_sub = fl.maximize(budget = int(n_sup), optimizer = 'LazyGreedy', stopIfZeroGain = False, stopIfNegativeGain = False, verbose = False)
	indices = np.array([i[0] for i in sim_sub])

	return np.sort(indices), data

def sup_subset_indices(path_json, path_pkl, n_sup, qc = 0.85):
	'''
		A function for supervised subset selection(the subset to be labeled) whcih just returns indices
	
	Args:
		path_json: Path to json file of number to string(class name) map
		path_pkl: Path to the pickle file containing all the training data in standard format
		n_sup: Number of instances to be found during supervised subset selection
		qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85

	Return:
		numpy.ndarray of indices(shape is (n_sup,), each element lies in [0,num_instances)), the result of subset selection
	'''
	indices, _ = sup_subset(path_json, path_pkl, n_sup, qc)

	return indices

def sup_subset_save_files(path_json, path_pkl, path_save_L, path_save_U, n_sup, qc = 0.85):
	'''
		A function for supervised subset selection(the subset to be labeled) which makes separate pickle files of data, one for those to be labelled, other that can be left unlabelled
	
	Args:
		path_json: Path to json file of number to string(class name) map
		path_pkl: Path to the pickle file containing all the training data in standard format
		path_save_L: Path to save the pickle file of set of instances to be labelled. Note that instances are not labelled yet. Extension should be .pkl
		path_save_U: Path to save the pickle file of set of instances that can be left unlabelled. Extension should be .pkl
		n_sup: number of instances to be found during supervised subset selection
		qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85

	Return:
		numpy.ndarray of indices(shape is (n_sup,), each element lies in [0,num_instances)), the result of subset selection. Also two pickle files are saved at path_save_L and path_save_U
	'''
	indices, data = sup_subset(path_json, path_pkl, n_sup, qc)

	false_mask = np.ones(data[0].shape[0], dtype = bool)
	false_mask[indices] = False

	save_file_L = open(path_save_L, 'wb')
	save_file_U = open(path_save_U, 'wb')

	for i in range(10):
		if i < 7:
			if data[i].shape[0] == 0:
				pickle.dump(data[i], save_file_L)
				pickle.dump(data[i], save_file_U)
			else:
				pickle.dump(data[i][indices], save_file_L)
				pickle.dump(data[i][false_mask], save_file_U)
		elif i >= 7:
			pickle.dump(data[i], save_file_L)
			pickle.dump(data[i], save_file_U)

	save_file_L.close()
	save_file_U.close()

	return indices


def replace_in_pkl(path, path_save, np_array, index):
	'''
		A function to insert the true labels, after labeling the instances, to the pickle file

	Args:
		path: Path to the pickle file containing all the data in standard format
		path_save: Path to save the pickle file after replacing the 'L'(true labels numpy array) of data in path pickle file
		np_array: The data which is to be used to replace the data in path pickle file with
		index: Index of the numpy array, in data of path pickle file, to be replaced with np_array. Value should be in [0,8]
	Return:
		No return value. A pickle file is generated at path_save
	'''
	assert type(index) == np.int and index >=0 and index < 9
	assert check_path.exists(path) #path is imported from os above
	data = []
	with open(path, 'rb') as file:
		for i in range(9):
			data.append(pickle.load(file))
			assert type(data[i]) == np.ndarray
		data.append(pickle.load(file))

	assert data[index].shape[0] == 0 or np_array.shape == data[index].shape

	save_file = open(path_save, 'wb')
	for i in range(10):
		if i == index:
			pickle.dump(np_array, save_file)
		else:
			pickle.dump(data[i], save_file)
	save_file.close()

	return
	
def insert_true_labels(path, path_save, labels):
	'''
		A function to insert the true labels, after labeling the instances, to the pickle file

	Args:
		path: Path to the pickle file containing all the data in standard format
		path_save: Path to save the pickle file after replacing the 'L'(true labels numpy array) of data in path pickle file
		labels: The true labels of the data in pickle file. numpy.ndarray of shape (num_instances, 1)

	Return:
		No return value. A pickle file is generated at path_save
	'''
	assert check_path.exists(path) #path is imported from os above
	data = []
	with open(path, 'rb') as file:
		for i in range(9):
			data.append(pickle.load(file))
			assert type(data[i]) == np.ndarray
		data.append(pickle.load(file))

	assert labels.shape[0] == data[0].shape[0] and labels.shape[1] == 1

	save_file = open(path_save, 'wb')
	for i in range(10):
		if i == 3:
			pickle.dump(labels, save_file)
		elif i == 4:
			pickle.dump(np.ones([labels.size ,1]), save_file)
		else:
			pickle.dump(data[i], save_file)
	save_file.close()

	return


import numpy as np
import enum, json
from typing import Optional

from ..lf_set import LFSet
from ..apply import LFApplier
from ..analysis import LFAnalysis
from ..lf import ABSTAIN 
from ..utils import dump_to_pickle
from ..data_types import DataPoints


class PreLabels:
    """Generate noisy lables, continuous score  from lf's applied on data  

    Args:
        name (str): Name for this object.
        data (DataPoints): Datapoints.
        gold_labels (Optional[DataPoints]): Labels for datapoints if available.
        rules (LFSet): Set of Rules to generate noisy labels for the dataset.
        exemplars (DataPoints): [description]
    """    
    def __init__(
        self,
        name: str,
        data: DataPoints,
        rules: LFSet,
        num_classes: int,
        labels_enum,
        data_feats: Optional[DataPoints] = np.array([]),
        gold_labels: Optional[DataPoints] = np.array([]),
        exemplars: DataPoints=np.array([]),
    ) -> None:       
        """Instantiates PreLabels class with dataset and set of LFs to noisily label the dataset
        """
        self.name = name
        self._data = data
        self._rules = rules
        self._num_classes = num_classes
        self._labels_enum = labels_enum
        self._data_feats = data_feats
        self._gold_labels = gold_labels
        self._R = exemplars
        self._L = None
        self._S = None

        assert num_classes == len(labels_enum) 
        
        lab_vals = set(item.value for item in self._labels_enum)
        assert len(lab_vals)==self._num_classes 
        
        lab_nams = set(item.name for item in self._labels_enum)
        assert 'ABSTAIN' not in lab_nams

        assert (self._data_feats.shape[0]==0) or (self._data_feats.shape[0]==self._data.shape[0])
        assert (len(self._gold_labels)==self._data.shape[0]) or (self._gold_labels.shape[0]==0)
        unique_labs = set(np.unique(self._gold_labels))
        assert unique_labs.issubset(lab_vals)
        assert (self._R.shape[0]==0) or (self._R.shape[0]==self._data.shape[0] and self._R.shape[1]==len(self._rules))


    def get_labels(self):
        """Applies LFs to the dataset to generate noisy labels and returns noisy labels and confidence scores

        Returns:
            Tuple(DataPoints, DataPoints): Noisy Labels and Confidences
        """
        if self._L is None or self._L is None:
            applier = LFApplier(lf_set = self._rules)
            L,S = applier.apply(self._data)
            self._L = L
            self._S = S
        return self._L, self._S

    def analyse_lfs(self,plot=False):
        """Analyse the lfs in LFSet on data

        Args:
            plot (bool, optional): Plot the values. Defaults to False.

        Returns:
            DataFrame: dataframe consisting of Ploarity, Coverage, Overlap, Conflicts, Empirical Acc
        """        
        if self._L is None or self._L is None:
            applier = LFApplier(lf_set = self._rules)
            L,S = applier.apply(self._data)
            self._L = L
            self._S = S
        
        analysis = LFAnalysis(self._labels_enum,self._L,self._rules)
        if len(self._gold_labels) == 0:
            df = analysis.lf_summary(plot=plot)
        else:
            df = analysis.lf_summary(self._gold_labels,plot=plot)
        return df
        

    def generate_json(self, filename=None):
        """Generates a json file with label value to label name mapping

        Args:
            filename (str, optional): Name for json file. Defaults to None.
        """
        if filename is None:
            filename = self.name+"_json.json"
        
        dic = {}
        for e in self._labels_enum:
            dic[e.value]=e.name

        with open(filename, "w") as outfile:
            json.dump(dic, outfile)
        
    def generate_pickle(self, filename=None):
        """Generates a pickle file with noisy labels, confidence and other Metadata

        Args:
            filename (str, optional): Name for pickle file. Defaults to None.
        """
        if filename is None:
            filename = self.name+"_pickle.pkl"
        
        if (self._L is None or self._S is None):
            applier = LFApplier(lf_set = self._rules)
            L,S = applier.apply(self._data)
            self._L = L
            self._S = S

        num_inst=self._data.shape[0]
        num_rules=self._L.shape[1]

        x=self._data_feats
        l=self._L

        m=(self._L!=ABSTAIN).astype(int)                                        # lf covers example or not 
        L=self._gold_labels                                                     # true labels
        L = L.reshape(L.size,1)
        d=np.ones((num_inst, 1))                                                # belongs to labeled data or not
        r=self._R                                                               # exemplars

        s=self._S                                                               # continuous scores
        n=np.array([lf._is_cont for lf in self._rules.get_lfs()])               # lf continuous or not
        k=np.array([lf._label.value for lf in self._rules.get_lfs()])           # lf associated to which class

        output = [x,l,m,L,d,r,s,n,k,self._num_classes]
        dump_to_pickle(filename, output)









import torch
from torch import optim
import pickle
from os import path as check_path
import numpy as np
from tqdm import tqdm
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score

from ..utils.data_editor import get_data, get_classes, get_predictions
from ..utils.utils_cage import probability, log_likelihood_loss, precision_loss, predict_gm_labels

class Cage:
	'''
	Cage class:
		Class for Data Programming using CAGE
		[Note: from here on, graphical model(gm) and CAGE algorithm terms are used interchangeably]

	Args:
		path_json: Path to json file consisting of number to string(class name) map
		n_lfs: number of labelling functions used to generate pickle files

	'''
	def __init__(self, path_json, n_lfs):
		assert type(path_json) == str
		assert type(n_lfs) == np.int or type(n_lfs) == np.float

		use_cuda = torch.cuda.is_available()
		self.device = torch.device("cuda" if use_cuda else "cpu")
		torch.backends.cudnn.benchmark = True

		self.class_dict = get_classes(path_json)
		self.class_list = list((self.class_dict).keys())
		self.class_list.sort()
		self.n_classes = len(self.class_dict)

		self.class_map = {value: index for index, value in enumerate(self.class_list)}
		self.class_map[None] = self.n_classes

		self.n_lfs = int(n_lfs)
		self.n, self.k = None, None #continuous_mask, labels of LFs

		self.pi = torch.ones((self.n_classes, self.n_lfs), device = self.device).double()
		(self.pi).requires_grad = True

		self.theta = torch.ones((self.n_classes, self.n_lfs), device = self.device).double()
		(self.theta).requires_grad = True

	def save_params(self, save_path):
		'''
			member function to save parameters of Cage

		Args:
			save_path: path to pickle file to save parameters
		'''
		file_ = open(save_path, 'wb')
		pickle.dump(self.theta, file_)
		pickle.dump(self.pi, file_)
		pickle.dump(self.n_classes, file_)
		pickle.dump(self.n_lfs, file_)
		file_.close()
		return

	def load_params(self, load_path):
		'''
			member function to load parameters to Cage

		Args:
			load_path: path to pickle file to load parameters
		'''
		assert check_path.exists(load_path)
		file_ = open(load_path, 'rb')
		self.theta = pickle.load(file_)
		self.pi = pickle.load(file_)
		assert self.n_classes == pickle.load(file_)
		assert self.n_lfs == pickle.load(file_)
		file_.close()

		assert (self.pi).shape == (self.n_classes, self.n_lfs)
		assert (self.theta).shape == (self.n_classes, self.n_lfs)
		return

	def fit_and_predict_proba(self, path_pkl, path_test = None, path_log = None, qt = 0.9, qc = 0.85, metric_avg = ['binary'], n_epochs = 100, lr = 0.01):
		'''
		Args:
			path_pkl: Path to pickle file of input data in standard format
			path_test: Path to the pickle file containing test data in standard format
			path_log: Path to log file. No log is produced if path_test is None. Default is None which prints accuracies/f1_scores is printed to terminal
			qt: Quality guide of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.9
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			metric_avg: List of average metric to be used in calculating f1_score, default is ['binary']. Use None for not calculating f1_score
			n_epochs:Number of epochs, default is 100
			lr: Learning rate for torch.optim, default is 0.01

		Return:
			numpy.ndarray of shape (num_instances, num_classes) where i,j-th element is the probability of ith instance being the jth class(the jth value when sorted in ascending order of values in Enum)
		'''
		assert  type(path_pkl) == str
		assert (type(qt) == np.float and (qt >= 0 and qt <= 1)) or (type(qt) == np.ndarray and (np.all(np.logical_and(qt>=0, qt<=1)) ) )\
		 or (type(qt) == np.int and (qt == 0 or qt == 1))

		assert (type(qc) == np.float and (qc >= 0 and qc <= 1)) or (type(qc) == np.ndarray and (np.all(np.logical_and(qc>=0, qc<=1)) ) )\
		 or (type(qc) == np.int and (qc == 0 or qc == 1))

		for temp in metric_avg:
			assert temp in ['micro', 'macro', 'samples','weighted', 'binary'] or metric_avg is None
		assert type(n_epochs) == np.int or type(n_epochs) == np.float
		assert type(lr) == np.int or type(lr) == np.float

		data = get_data(path_pkl, True, self.class_map)
		m = torch.abs(torch.tensor(data[2], device = self.device).long())
		s = torch.tensor(data[6], device = self.device).double() # continuous score
		if self.n == None:	
			self.n = torch.tensor(data[7], device = self.device).double() # Mask for s/continuous_mask
		else:
			assert torch.all(torch.tensor(data[7], device = self.device).double().eq(self.n))
		if self.k == None:
			self.k = torch.tensor(data[8], device = self.device).long() # LF's classes
		else:
			assert torch.all(torch.tensor(data[8], device = self.device).long().eq(self.k))
		s[s > 0.999] = 0.999 # clip s
		s[s < 0.001] = 0.001 # clip s

		assert self.n_lfs == m.shape[1]
		assert self.n_classes == data[9]

		qt_ = torch.tensor(qt, device = self.device).double() if type(qt) == np.ndarray else (torch.ones(self.n_lfs, device = self.device).double() * qt)
		qc_ = torch.tensor(qc, device = self.device).double() if type(qc) == np.ndarray else qc
		metric_avg_ = list(set(metric_avg))
		n_epochs_ = int(n_epochs)
		
		optimizer = optim.Adam([self.theta, self.pi], lr=lr, weight_decay=0)

		file = None
		if path_test != None and path_log != None:
			file = open(path_log, "a+")
			file.write("CAGE log:\tn_classes: {}\tn_LFs: {}\tn_epochs: {}\tlr: {}\n".format(self.n_classes, self.n_lfs, n_epochs, lr))
		elif path_test != None:
			print("CAGE log:\tn_classes: {}\tn_LFs: {}\tn_epochs: {}\tlr: {}".format(self.n_classes, self.n_lfs, n_epochs, lr))

		y_true_test = None
		s_test, m_test = None, None
		if path_test != None:
			data = get_data(path_test, True, self.class_map)
			m_test, y_true_test, s_test = data[2], data[3], data[6]
			assert m_test.shape[0] == y_true_test.shape[0]
			y_true_test = y_true_test.flatten()
			assert self.n_lfs == m_test.shape[1]
			assert self.n_classes == data[9]
			assert torch.all(torch.tensor(data[7], device = self.device).double().eq(self.n))
			assert torch.all(torch.tensor(data[8], device = self.device).long().eq(self.k))

		assert np.all(np.logical_and(y_true_test >= 0, y_true_test < self.n_classes))

		with tqdm(total=n_epochs_) as pbar:
			for epoch in range(n_epochs_):
				optimizer.zero_grad()
				loss = log_likelihood_loss(self.theta, self.pi, m, s, self.k, self.n_classes, self.n, qc_, self.device)
				prec_loss = precision_loss(self.theta, self.k, self.n_classes, qt_, self.device)
				loss += prec_loss

				if path_test != None:
					y_pred = self.__predict_specific(m_test, s_test, qc_)
					if path_log != None:
						file.write("Epoch: {}\ttest_accuracy_score: {}\n".format(epoch, accuracy_score(y_true_test, y_pred)))
					else:
						print("Epoch: {}\ttest_accuracy_score: {}".format(epoch, accuracy_score(y_true_test, y_pred)))
					if epoch == n_epochs_-1:
						print("final_test_accuracy_score: {}".format(accuracy_score(y_true_test, y_pred)))
					for temp in metric_avg_:
						if path_log != None:
							file.write("Epoch: {}\ttest_average_metric: {}\ttest_f1_score: {}\n".format(epoch, temp, f1_score(y_true_test, y_pred, average = temp)))
						else:
							print("Epoch: {}\ttest_average_metric: {}\ttest_f1_score: {}".format(epoch, temp, f1_score(y_true_test, y_pred, average = temp)))
						if epoch == n_epochs_-1:
							print("test_average_metric: {}\tfinal_test_f1_score: {}".format(temp, f1_score(y_true_test, y_pred, average = temp)))

				loss.backward()
				optimizer.step()
				pbar.update()

		if path_test != None and path_log != None:
			file.close()

		return (probability(self.theta, self.pi, m, s, self.k, self.n_classes, self.n, qc_, self.device)).cpu().detach().numpy()

	def fit_and_predict(self, path_pkl, path_test = None, path_log = None, qt = 0.9, qc = 0.85, metric_avg = ['binary'], n_epochs = 100, lr = 0.01, need_strings = False):
		'''
		Args:
			path_pkl: Path to pickle file of input data in standard format
			path_test: Path to the pickle file containing test data in standard format
			path_log: Path to log file. No log is produced if path_test is None. Default is None which prints accuracies/f1_scores is printed to terminal
			qt: Quality guide of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.9
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			metric_avg: List of average metric to be used in calculating f1_score, default is ['binary']
			n_epochs:Number of epochs, default is 100
			lr: Learning rate for torch.optim, default is 0.01
			need_strings: If True, the output will be in the form of strings(class names). Else it is in the form of class values(given to classes in Enum). Default is False

		Return:
			numpy.ndarray of shape (num_instances,) which are aggregated/predicted labels. Elements are numbers/strings depending on need_strings attribute is false/true resp.
		'''
		assert type(need_strings) == np.bool
		proba = self.fit_and_predict_proba(path_pkl, path_test, path_log, qt, qc, metric_avg, n_epochs, lr)
		return get_predictions(proba, self.class_map, self.class_dict, need_strings)

	def __predict_specific(self, m_test, s_test, qc_):
		'''
			Used to predict labels based on s_test and m_test

		Args:
			m_test: numpy arrays of shape (num_instances, num_rules), m_test[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
			s_test: numpy arrays of shape (num_instances, num_rules), s_test[i][j] is the continuous score of jth LF on ith instance
			qc_: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1
		
		Return:
			numpy.ndarray of shape (num_instances,) which are predicted labels. Note that here the class labels appearing may not be the ones used in the Enum
			[Note: no aggregration/algorithm-running will be done using the current input]
		'''
		s_temp = torch.tensor(s_test, device = self.device).double()
		s_temp[s_temp > 0.999] = 0.999
		s_temp[s_temp < 0.001] = 0.001
		assert m_test.shape == s_test.shape
		assert m_test.shape[1] == self.n_lfs
		assert np.all(np.logical_or(m_test == 1, m_test == 0))
		m_temp = torch.abs(torch.tensor(m_test, device = self.device).long())
		return predict_gm_labels(self.theta, self.pi, m_temp, s_temp, self.k, self.n_classes, self.n, qc_, self.device)

	def predict_proba(self, path_test, qc = 0.85):
		'''
			Used to predict labels based on a pickle file with path path_test

		Args:
			path_test: Path to the pickle file containing test data set in standard format
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85

		Return:
			numpy.ndarray of shape (num_instances, num_classes) where i,j-th element is the probability of ith instance being the jth class(the jth value when sorted in ascending order of values in Enum)
			[Note: no aggregration/algorithm-running will be done using the current input]
		'''
		assert (type(qc) == np.float and (qc >= 0 and qc <= 1)) or (type(qc) == np.ndarray and (np.all(np.logical_and(qc>=0, qc<=1)) ) )\
		 or (type(qc) == np.int and (qc == 0 or qc == 1))
		data = get_data(path_test, True, self.class_map)
		assert (data[2]).shape[1] == self.n_lfs and data[9] == self.n_classes
		temp_k = torch.tensor(data[8], device = self.device).long()
		assert self.k == None or torch.all(temp_k.eq(self.k))
		temp_n = torch.tensor(data[7], device = self.device).double()
		assert self.n == None or torch.all(temp_n.eq(self.n))
		s_test = torch.tensor(data[6], device = self.device).double()
		s_test[s_test > 0.999] = 0.999
		s_test[s_test < 0.001] = 0.001
		m_test = torch.abs(torch.tensor(data[2], device = self.device).long())	

		qc_ = torch.tensor(qc).double() if type(qc) == np.ndarray else qc
		if self.n == None or self.k == None:
			print("Warning: Predict is used before training any paramters in Cage class. Hope you have loaded parameters.")
		return (probability(self.theta, self.pi, m_test, s_test, temp_k, self.n_classes, temp_n, qc_, self.device)).cpu().detach().numpy()
		
	def predict(self, path_test, qc = 0.85, need_strings = False):
		'''
			Used to predict labels based on a pickle file with path path_test
			
		Args:
			path_test: Path to the pickle file containing test data set in standard format
			qc: Quality index of shape (n_lfs,) of type numpy.ndarray OR a float. Values must be between 0 and 1. Default is 0.85
			need_strings: If True, the output will be in the form of strings(class names). Else it is in the form of class values(given to classes in Enum). Default is False

		Return:
			numpy.ndarray of shape (num_instances,) which are predicted labels. Elements are numbers/strings depending on need_strings attribute is false/true resp.
			[Note: no aggregration/algorithm-running will be done using the current input]
		'''
		assert type(need_strings) == np.bool
		return get_predictions(self.predict_proba(path_test, qc), self.class_map, self.class_dict, need_strings)


import collections

f_d = 'f_d'
f_d_U = 'f_d_U'
test_w = 'test_w'

train_modes = [f_d, f_d_U]

F_d_U_Data = collections.namedtuple('GMMDataF_d_U', 'x l m L d r')
F_d_Data = collections.namedtuple('GMMDataF_d', 'x labels')


# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()

# implementation of generalizer cross entropy loss 
# eq6 of paper https://arxiv.org/pdf/1805.07836.pdf 


# loss for multiclass prediction problem 
# (logits corresponding to a softmax distribution and corresponding labels)
def generalized_cross_entropy(logits, one_hot_labels,q=0.6):
    '''
    Func Desc:
    Computes the generalized cross entropy loss

    Input:
    logits([batch_size, num_classes]) - weights
    one_hot_labels([batch_size, num_classes])
    q (default = 0.6)

    Output:
    loss

    '''
    #logits: [batch_size, num_classes]
    #one_hot_labels: [batch_size, num_classes]
    if q == 0.0:
        #for q=0 in limit, this is usual cross entropy
        loss = tf.nn.softmax_cross_entropy_with_logits(labels=one_hot_labels,logits=logits)
        loss = tf.reduce_mean(loss)
    else:
        exp_logits = tf.exp(logits)
        normalizer = tf.reduce_sum(exp_logits,axis=-1)#,keep_dims=True)
        normalizer_q = tf.pow(normalizer,q)
        exp_logits_q = tf.exp(logits*q)
        f_j_q = exp_logits_q / normalizer_q
        loss = (1 - f_j_q)/q
        loss = tf.reduce_sum(loss * one_hot_labels, axis=-1)
        loss = tf.reduce_mean(loss)
    return loss

# loss for a particular class
# p = probability of that class 
def generalized_cross_entropy_bernoulli(p,q=0.2):
    '''
    Func Desc:
    computes the bernoulli generalized cross entropy

    Input:
    p - base
    q (default = 0.2) - exponent

    Output:
    loss
    '''
    return (1 - tf.pow(p,q))/q





'''
The common utils to CAGE and JL algorithms are in this file. Don't change the name or location of this file.
'''

import numpy as np 
import torch
from torch.distributions.beta import Beta

def phi(theta, l, device):
	'''
		Graphical model utils: A helper function

	Args:
		theta: [n_classes, n_lfs], the parameters
		l: [n_lfs]
		device: 'cuda' if drivers are available, else 'cpu'

	Return:
		a tensor of shape [n_classes, n_lfs], element wise product of input tensors(each row of theta dot product with l)
	'''
	return theta * torch.abs(l).double().to(device=device)


def calculate_normalizer(theta, k, n_classes, device):
	'''
		Graphical model utils: Used to find Z(the normaliser) in CAGE. Eq(4) in :cite:p:`2020:CAGE`

	Args:
		theta: [n_classes, n_lfs], the parameters
		k: [n_lfs], labels corresponding to LFs
		n_classes: num of classes/labels
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a real value, representing the normaliser
	'''
	z = 0
	for y in range(n_classes):
		m_y = torch.exp(phi(theta[y], torch.ones(k.shape), device))
		z += (1 + m_y).prod()
	return z


def probability_l_y(theta, m, k, n_classes, device):
	'''
		Graphical model utils: Used to find probability involving the term psi_theta(in Eq(1) in :cite:p:`2020:CAGE`), the potential function for all LFs

	Args:
		theta: [n_classes, n_lfs], the parameters
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a tensor of shape [n_instances, n_classes], the psi_theta value for each instance, for each class(true label y)
	'''
	probability = torch.zeros((m.shape[0], n_classes), device = device)
	z = calculate_normalizer(theta, k, n_classes, device)
	for y in range(n_classes):
		probability[:, y] = torch.exp(phi(theta[y], m, device).sum(1)) / z

	return probability.double()


def probability_s_given_y_l(pi, s, y, m, k, continuous_mask, qc):
	'''
		Graphical model utils: Used to find probability involving the term psi_pi(in Eq(1) in :cite:p:`2020:CAGE`), the potential function for all continuous LFs

	Args:
		pi: [n_lfs], the parameters for the class y
		s: [n_instances, n_lfs], s[i][j] is the continuous score of ith instance given by jth continuous LF
		y: a value in [0, n_classes-1], representing true label, for which psi_pi is calculated
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		continuous_mask: [n_lfs], continuous_mask[i] is 1 if ith LF has continuous counter part, else it is 0
		qc: a float value OR [n_lfs], qc[i] quality index for ith LF. Value(s) must be between 0 and 1
	
	Return:
		a tensor of shape [n_instances], the psi_pi value for each instance, for the given label(true label y)
	'''
	eq = torch.eq(k.view(-1, 1), y).double().t()
	r = qc * eq.squeeze() + (1 - qc) * (1 - eq.squeeze())
	params = torch.exp(pi)
	probability = 1
	for i in range(k.shape[0]):
		temp = Beta(r[i] * params[i], params[i] * (1 - r[i]))
		probability *=  torch.exp(temp.log_prob(s[:, i].double())) * m[:, i].double() * continuous_mask[i] \
				 + (1 - m[:, i]).double() + (1 - continuous_mask[i])
	return probability


def probability(theta, pi, m, s, k, n_classes, continuous_mask, qc, device):
	'''
		Graphical model utils: Used to find probability of given instances for all possible true labels(y's). Eq(1) in :cite:p:`2020:CAGE`

	Args:
		theta: [n_classes, n_lfs], the parameters
		pi: [n_classes, n_lfs], the parameters
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		s: [n_instances, n_lfs], s[i][j] is the continuous score of ith instance given by jth continuous LF
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		continuous_mask: [n_lfs], continuous_mask[i] is 1 if ith LF has continuous counter part, else it is 0
		qc: a float value OR [n_lfs], qc[i] quality index for ith LF. Value(s) must be between 0 and 1
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a tensor of shape [n_instances, n_classes], the probability for an instance being a particular class
	'''
	p_l_y = probability_l_y(theta, m, k, n_classes, device)
	p_s = torch.ones(s.shape[0], n_classes, device = device).double()
	for y in range(n_classes):
		p_s[:, y] = probability_s_given_y_l(pi[y], s, y, m, k, continuous_mask, qc)
	return p_l_y * p_s


def log_likelihood_loss(theta, pi, m, s, k, n_classes, continuous_mask, qc, device):
	'''
		Graphical model utils: Negative of log likelihood loss. Negative of Eq(6) in :cite:p:`2020:CAGE`

	Args:
		theta: [n_classes, n_lfs], the parameters
		pi: [n_classes, n_lfs], the parameters
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		s: [n_instances, n_lfs], s[i][j] is the continuous score of ith instance given by jth continuous LF
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		continuous_mask: [n_lfs], continuous_mask[i] is 1 if ith LF has continuous counter part, else it is 0
		qc: a float value OR [n_lfs], qc[i] quality index for ith LF. Value(s) must be between 0 and 1
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a real value, negative of summation over (the log of probability for an instance, marginalised over y(true labels))
	'''
	eps = 1e-8
	return - torch.log(probability(theta, pi, m, s, k, n_classes, continuous_mask, qc, device).sum(1) + eps).sum() / s.shape[0]


def precision_loss(theta, k, n_classes, a, device): 
	'''
		Graphical model utils: Negative of the regularizer term in Eq(9) in :cite:p:`2020:CAGE`

	Args:
		theta: [n_classes, n_lfs], the parameters
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		a: [n_lfs], a[i] is the quality guide for ith LF. Value(s) must be between 0 and 1
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		a real value, negative of regularizer term
	'''
	n_lfs = k.shape[0]
	prob = torch.ones(n_lfs, n_classes, device = device).double()
	z_per_lf = 0
	for y in range(n_classes):
		m_y = torch.exp(phi(theta[y], torch.ones(n_lfs, device = device), device))
		per_lf_matrix = torch.tensordot((1 + m_y).view(-1, 1), torch.ones(m_y.shape, device = device).double().view(1, -1), 1) \
		- torch.eye(n_lfs, device = device).double()
		prob[:, y] = per_lf_matrix.prod(0).double()
		z_per_lf += prob[:, y].double()
	prob /= z_per_lf.view(-1, 1)
	correct_prob = torch.zeros(n_lfs, device = device)
	for i in range(n_lfs):
		correct_prob[i] = prob[i, k[i]]
	loss = a * torch.log(correct_prob).double() + (1 - a) * torch.log(1 - correct_prob).double()
	return -loss.sum()

def predict_gm_labels(theta, pi, m, s, k, n_classes, continuous_mask, qc, device):
	'''
		Graphical model utils: Used to predict the labels after the training is done

	Args:
		theta: [n_classes, n_lfs], the parameters
		pi: [n_classes, n_lfs], the parameters
		m: [n_instances, n_lfs], m[i][j] is 1 if jth LF is triggered on ith instance, else it is 0
		s: [n_instances, n_lfs], s[i][j] is the continuous score of ith instance given by jth continuous LF
		k: [n_lfs], k[i] is the class of ith LF, range: 0 to num_classes-1
		n_classes: num of classes/labels
		continuous_mask: [n_lfs], continuous_mask[i] is 1 if ith LF has continuous counter part, else it is 0
		qc: a float value OR [n_lfs], qc[i] quality index for ith LF. Value(s) must be between 0 and 1
		device: 'cuda' if drivers are available, else 'cpu'
	
	Return:
		numpy.ndarray of shape (n_instances,), the predicted class for an instance
	'''
	return np.argmax(probability(theta, pi, m, s, k, n_classes, continuous_mask, qc, device).cpu().detach().numpy(), 1)


from typing import Any, Callable, List, Set, Mapping, Optional
import warnings
import pickle
import numpy as np

from ..data_types import DataPoint, DataPoints
from ..lf import LabelingFunction

class LFSet:              # LFSet
    """Class for Set of Labeling Functions

    Args:
        name (str): Name for this LFset.
        lfs (List[LabelingFunction], optional): List of LFs to add to this object. Defaults to [].
    """
    def __init__(
        self,
        name: str,
        lfs: List[LabelingFunction] = [],
    ) -> None:
        """Instantiates LFSet class with list of labeling functions      
        """
        self.name = name
        self._lfs = lfs

    def get_lfs(
        self,
    ) -> Set[LabelingFunction]:
        """Returns LFs contained in this LFSet object

        Returns:
            Set[LabelingFunction]: LFs in this LFSet
        """
        return self._lfs
        
    def add_lf(
        self,
        lf: LabelingFunction,
    ) -> None:
        """Adds single LF to this LFSet

        Args:
            lf (LabelingFunction): LF to add
        """
        if lf in self._lfs:
            warnings.warn('Attempting to add duplicate LF to LFset') 
        else:
            self._lfs.append(lf)
    
    def add_lf_list(
        self,
        lf_list: List[LabelingFunction],
    ) -> None:
        """Adds a list of LFs to this LFSet

        Args:
            lf_list (List[LabelingFunction]): List of LFs to add to this LFSet
        """
        if len(set(self._lfs).intersection(lf_list))>0:
            warnings.warn('Attempting to add duplicate LF to LFset') 
        self._lfs = self._lfs + lf_list
    
    def remove_lf(
        self,
        lf: LabelingFunction
    ) -> None:
        """Removes a LF from this set

        Args:
            lf (LabelingFunction): LF to remove from this set

        Raises:
            Warning: If LF not already in LFset
        """
        if lf in self._lfs:
            self._lfs.remove(lf)
        else:
            warnings.warn("Trying to remove an LF not in this LF set!")
    



import warnings
import numpy as np
from itertools import product
import scipy.sparse as sparse
from collections import OrderedDict
from pandas import DataFrame, Series
from typing import List, Optional, Union
from sklearn.metrics import confusion_matrix

from ..utils import plot_df_bar
from ..lf import LabelingFunction

warnings.simplefilter(action='ignore', category=FutureWarning)

class LFAnalysis:   
    """Run analysis on LFs using label matrix.

        Args:
            L (np.ndarray): Label matrix where L_{i,j} is the label given by the jth LF to the ith x instance
            lfs (Optional[List[LabelingFunction]], optional): Labeling functions used to generate `'L``. Defaults to None.
            abstain (int, optional): label associated with abstain. Defaults to -1.

        Raises:
            ValueError:  If number of LFs and number of LF matrix columns differ    
    """

    def __init__(
        self, enum, L: np.ndarray, rules = None
    ) -> None:  
        self.L,self.mapping = self._create_L(enum,L)
        self._L_sparse = sparse.csr_matrix(self.L + 1)
        self._lf_names = None
        lfs = list(rules.get_lfs())
        if lfs is not None:
            if len(lfs) != self.L.shape[1]:
                raise ValueError(
                    f"Number of LFs ({len(lfs)}) and number of "
                    f"LF matrix columns ({self.L.shape[1]}) are different"
                )
            self._lf_names = [lf.name for lf in lfs]

    def _create_L(self,enum,L):
        """Map the enum values to non-ve integers and abstain to -1"""        
        mapping = {}
        j = 0
        L_num = -1 * np.ones_like(L,dtype=float)
        for i in enum:
            mapping[i.value] = j
            L_num[L==i.value] = j
            j = j+1
        return L_num, mapping
        


    def _covered_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled by at least one LF."""
        return np.ravel(np.where(self._L_sparse.sum(axis=1) != 0, 1, 0))
    
    def _overlapped_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled by more than one LF."""
        return np.where(np.ravel((self._L_sparse != 0).sum(axis=1)) > 1, 1, 0)

    
    def _conflicted_data_points(self) -> np.ndarray:
        """Get indicator vector z where z_i = 1 if x_i is labeled differently by two LFs."""
        m = sparse.diags(np.ravel(self._L_sparse.max(axis=1).todense()))
        return np.ravel(
            np.max(m @ (self._L_sparse != 0) != self._L_sparse, axis=1)
            .astype(int)
            .todense()
        )

    
    def label_coverage(self) -> float:
        """Compute the fraction of data points with at least one label.

        Returns:
            float: Fraction of data points with labels
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).label_coverage()
            0.8
        """
        return self._covered_data_points().sum() / self.L.shape[0]
    

    def label_overlap(self) -> float:
        """Compute the fraction of data points with at least two (non-abstain) labels.

        Returns:
            float: Fraction of data points with overlapping labels
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).label_overlap()
            0.6  
        """
        return self._overlapped_data_points().sum() / self.L.shape[0]
    

    def label_conflict(self) -> float:
        """Compute the fraction of data points with conflicting (non-abstain) labels.

        Returns:
            float: Fraction of data points with conflicting labels
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).label_conflict()
            0.2
        """        
        return self._conflicted_data_points().sum() / self.L.shape[0]
    
    def lf_polarities(self) -> List[List[int]]:
        """Infer the polarities of each LF based on evidence in a label matrix.

        Returns:
            List[List[int]]: Unique output labels for each LF
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).lf_polarities()
            [[0, 1], [0], [0]]
        """
        return [
            sorted(list(set(self.L[:, i])))
            for i in range(self.L.shape[1])
        ]
    

    def lf_coverages(self) -> np.ndarray:
        """Compute frac. of examples each LF labels.

        Returns:
            np.ndarray: Fraction of labeled examples for each LF
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).lf_coverages()
            array([0.4, 0.8, 0.4])
        """
        return np.ravel((self.L != -1).sum(axis=0)) / self.L.shape[0]
    

    def lf_overlaps(self, normalize_by_coverage: bool = False) -> np.ndarray:
        """Compute frac. of examples each LF labels that are labeled by another LF.
        An overlapping example is one that at least one other LF returns a
        (non-abstain) label for.
        Note that the maximum possible overlap fraction for an LF is the LF's
        coverage, unless ``normalize_by_coverage=True``, in which case it is 1

        Args:
            normalize_by_coverage (bool, optional): Normalize by coverage of the LF,
                                                    so that it returns the percent of LF labels that have overlaps.
                                                    Defaults to False.

        Returns:
            np.ndarray: Fraction of overlapping examples for each LF
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).lf_overlaps()
            array([0.4, 0.6, 0.4])
            >>> LFAnalysis(L).lf_overlaps(normalize_by_coverage=True)
            array([1.  , 0.75, 1.  ])
        """
        overlaps = (
            (self._L_sparse != 0).T
            @ self._overlapped_data_points()
            / self._L_sparse.shape[0]
        )
        if normalize_by_coverage:
            overlaps /= self.lf_coverages()
        return np.nan_to_num(overlaps)
    

    def lf_conflicts(self, normalize_by_overlaps: bool = False) -> np.ndarray:
        """Compute frac. of examples each LF labels and labeled differently by another LF.
        A conflicting example is one that at least one other LF returns a
        different (non-abstain) label for.
        Note that the maximum possible conflict fraction for an LF is the LF's
        overlaps fraction, unless ``normalize_by_overlaps=True``, in which case it is 1.
        Parameters

        Args:
            normalize_by_overlaps (bool, optional): Normalize by overlaps of the LF, so that it returns the percent of LF
                                                    overlaps that have conflicts.
                                                    Defaults to False.

        Returns:
            np.ndarray: Fraction of conflicting examples for each LF
        
        Example:
            >>> L = np.array([
            ...     [-1, 0, 0],
            ...     [-1, -1, -1],
            ...     [1, 0, -1],
            ...     [-1, 0, -1],
            ...     [0, 0, 0],
            ... ])
            >>> LFAnalysis(L).lf_conflicts()
            array([0.2, 0.2, 0. ])
            >>> LFAnalysis(L).lf_conflicts(normalize_by_overlaps=True)
            array([0.5       , 0.33333333, 0.        ])
        """        
        conflicts = (
            (self._L_sparse != 0).T
            @ self._conflicted_data_points()
            / self._L_sparse.shape[0]
        )
        if normalize_by_overlaps:
            conflicts /= self.lf_overlaps()
        return np.nan_to_num(conflicts)
    

    def lf_empirical_accuracies(self, Y: np.ndarray) -> np.ndarray:
        """Compute empirical accuracy against a set of labels Y for each LF.
        Usually, Y represents development set labels.

        Args:
            Y (np.ndarray): [n] np.ndarray of gold labels

        Returns:
            np.ndarray: Empirical accuracies for each LF
        """
        X = np.where(
            self.L == -1,
            0,
            np.where(self.L == np.vstack([Y] * self.L.shape[1]).T, 1, -1),
        )
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.nan_to_num(0.5 * (X.sum(axis=0) / (self.L != -1).sum(axis=0) + 1))
    

    # def lf_empirical_probs(self, Y: np.ndarray, k: int) -> np.ndarray:
    #     """Estimate conditional probability tables for each LF.
    #     Computes conditional probability tables, P(L | Y), for each LF using
    #     the provided true labels Y.

    #     Args:
    #         Y (np.ndarray):  The n-dim array of true labels
    #         k (int): The cardinality i.e. number of classes

    #     Returns:
    #         np.ndarray: An m x (k+1) x k np.ndarray representing the m (k+1) x k conditional probability
    #         tables P_i, where P_i[l,y] represents P(LF_i = l | Y = y) empirically calculated
    #     """        
    #     n, m = self.L.shape

    #     Y = Y#to_int_label_array(Y)

    #     # Compute empirical conditional probabilities
    #     # Note: Can do this more efficiently...
    #     P = np.zeros((m, k + 1, k))
    #     for y in range(k):
    #         is_y = np.where(Y == y, 1, 0)
    #         for j, l in product(range(m), range(-1, k)):
    #             P[j, l + 1, y] = np.where(self.L[:, j] == l, 1, 0) @ is_y / is_y.sum()
    #     return P
    

    def lf_summary(
        self, Y: Optional[np.ndarray] = None, plot: Optional[bool] = False
    ) -> DataFrame:
        """Create a pandas DataFrame with the various per-LF statistics.

        Args:
            Y (Optional[np.ndarray], optional): [n] np.ndarray of gold labels.
                                                If provided, the empirical accuracy for each LF will be calculated.
                                                Defaults to None.
            plot (Optional[bool], optional): If set to true a bar graph is plotted. Defaults to False.

        Returns:
            DataFrame: Summary statistics for each LF
        """
        n, m = self.L.shape
        lf_names: Union[List[str], List[int]]
        d: OrderedDict[str, Series] = OrderedDict()
        if self._lf_names is not None:
            lf_names = self._lf_names
        else:
            lf_names = list(range(m))

        # Remap the true labels values
        print('Y is ', Y)
        if Y is not None:
            Y = np.array([self.mapping[v] for v in Y])

        # Default LF stats
        d["Polarity"] = Series(data=self.lf_polarities(), index=lf_names)
        d["Coverage"] = Series(data=self.lf_coverages(), index=lf_names)
        d["Overlaps"] = Series(data=self.lf_overlaps(), index=lf_names)
        d["Conflicts"] = Series(data=self.lf_conflicts(), index=lf_names)

        if Y is not None:
            labels = np.unique(
                np.concatenate((Y.flatten(), self.L.flatten(), np.array([-1])))
            )
            confusions = [
                confusion_matrix(Y, self.L[:, i], labels=labels)[1:, 1:] for i in range(m)
            ]
            corrects = [np.diagonal(conf).sum() for conf in confusions]
            incorrects = [
                conf.sum() - correct for conf, correct in zip(confusions, corrects)
            ]
            accs = self.lf_empirical_accuracies(Y)
            d["Correct"] = Series(data=corrects, index=lf_names)
            d["Incorrect"] = Series(data=incorrects, index=lf_names)
            d["Emp. Acc."] = Series(data=accs, index=lf_names)
        
        data_frame = DataFrame(data=d, index=lf_names)
        
        if plot == True:
            plot_df_bar(data_frame,"seperate")

        return data_frame
    


import torch.nn as nn

class LogisticRegression(nn.Module):
	'''
		Class for Logistic Regression, used in Joint learning class/Algorithm

	Args:
		input_size: number of features
		output_size: number of classes
	'''
	def __init__(self, input_size, output_size):
		'''
		'''
		super(LogisticRegression, self).__init__()
		self.linear = nn.Linear(input_size, output_size)

	def forward(self, x):
		'''
		'''
		return self.linear(x)

class DeepNet(nn.Module):
	'''
		Class for Deep neural network, used in Joint learning class/Algorithm
		
	Args:
		input_size: number of features
		hidden_size: number of nodes in each of the two hidden layers
		output_size: number of classes
	'''
	def __init__(self, input_size, hidden_size, output_size):
		'''
		'''
		super(DeepNet, self).__init__()
		self.linear_1 = nn.Linear(input_size, hidden_size)
		self.linear_2 = nn.Linear(hidden_size, hidden_size)
		self.out = nn.Linear(hidden_size, output_size)

	def forward(self, x):
		'''
		'''
		out = nn.functional.relu(self.linear_1(x))
		out = nn.functional.relu(self.linear_2(out))
		return self.out(out)